from guppy.heapy.test import support
import io
import sys
import types
import unittest


class TestCase(support.TestCase):
    def setUp(self):
        support.TestCase.setUp(self)
        self.RefPat = self.heapy.RefPat
        self.iso = self.Use.iso

    def makegraph(self, width, length):
        # Generate a structure which will yield a high number
        # of shortest paths.
        # Returns a pair src, dst which are connected via a noncyclic graph
        # with many edges.
        # The length of each path (all shortest), number of edges will be length
        # The number of nodes will be 2 + width * (length - 1)
        # The number of paths will be
        #       width ** length, if width >= 1 and length >= 1

        dst = []
        ls = []
        for i in range(width):
            ls.append([dst])
        ls = [dst] * width
        for i in range(length-1):
            xs = []
            for j in range(width):
                ys = []
                xs.append(ys)
                for k in range(width):
                    ys.append(ls[k])
            ls = xs
        src = ls
        return src, dst

    def rp(self, dst, src, **kwds):
        iso = self.iso
        if src is not None:
            src = iso(src)
        rp = iso(dst).get_rp(src=src, **kwds)
        rp.mod.UniSet.summary_str.str_address = lambda a: '<address>'
        return rp


class RefPatCase(TestCase):
    def test_basic_methods(self):
        # Test basic methods: iteration, indexing, length, tree addressing via attribute access

        # Test iteration

        dst = src = []
        lists = [dst]
        for i in range(5):
            src = [src]
            lists.append(src)

        rp = self.rp(dst, src, depth=10)
        for i, x in enumerate(rp):
            if i < len(lists):
                self.asis(lists[i], x.theone)

        # Test indexing

        # First case, when already iterated over

        self.asis(rp[0].theone, lists[0])
        self.asis(rp[-2].theone, lists[-1])

        # Second case, when not iterated over before

        rp = self.rp(dst, src, depth=10)

        self.asis(rp[0].theone, lists[0])
        self.asis(rp[-2].theone, lists[-1])

        # Test length

        self.aseq(len(rp), len(lists) + 1)
        rp = self.rp(dst, src, depth=10)
        self.aseq(len(rp), len(lists) + 1)

        # Test attribute access

        self.asis(rp._.theone, lists[0])
        self.asis(rp.a.theone, lists[1])

        # Test attribute access, when not iterated over before

        rp = self.rp(dst, src, depth=10)
        self.asis(rp.a2.theone, lists[2])
        self.asis(rp.a.theone, lists[1])

        # Make sure attribute access is cached:
        # so it doesn't change when struct is changed

        lists[2].append(lists[0])
        rp.View.clear_retainers()
        rp.View.update_referrers(self.iso(lists[0]))
        self.asis(rp.a.theone, lists[1])

        # Test with recursive structure

        dst = []
        dst.append(dst)
        src = [dst]
        rp = self.rp(dst, src)
        self.asis(rp._.theone, dst)
        self.aseq(rp.a, self.iso(dst, src))
        self.aseq(rp.a, rp.a2)
        self.aseq(rp.a, rp[1])

    def test_presentation(self):

        output = io.StringIO()

        src = []

        def write(x):
            print(x, file=output)

        R = self.RefPat
        def test_pp(dst, src, result=None, **kwds):

            rp = self.rp(dst, src, **kwds)
            write(repr(rp))
            return rp

        dst = []
        src.append(dst)

        test_pp(dst, src)

        for i in range(5):
            x = dst
            dst = []
            x.append(dst)
        test_pp(dst, src)

        src, dst = self.makegraph(5, 7)

        test_pp(dst, src, depth=10)

        # Test that pp() prints limited number of lines

        src, dst = self.makegraph(5, 17)

        rp = test_pp(dst, src, depth=17)

        write(repr(rp.more))

        # Test more of more

        src, dst = self.makegraph(1, 30)

        rp = test_pp(dst, src, depth=35)

        m = rp.more

        write(repr(m))
        write(repr(m.more))
        m1 = m.more
        write(repr(m1))
        m2 = m.more
        write(repr(m2.more))
        write(str(m1.more))  # Test also that str() is the same as repr()

        # Test that we get back to start by .top

        write(m1.top)

        # Test that we get back to previous by .prev

        write(m1.prev)

        # Test that they won't say '...more lines...' if the # of lines is what is printed

        src, dst = self.makegraph(1, 30)

        rp = test_pp(dst, src, depth=10)

        # Test how no more lines is printed

        write(rp.more)
        write(rp.more.more)

        # Test that one more line is printed rather than '1 more line'

        src, dst = self.makegraph(1, 30)

        rp = test_pp(dst, src, depth=21)
        write(rp.more)

        # Test that we can do more without first printing

        rp = self.rp(dst, src, depth=20)

        write(rp.more)

        self.aseq(output.getvalue(), """\
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [R] 1 tuple: <address>*1
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [-] 1 list: <address>*1
 3: a3       [-] 1 list: <address>*1
 4: a4 ------ [-] 1 list: <address>*1
 5: a5         [-] 1 list: <address>*1
 6: a6 -------- [-] 1 list: <address>*1
 7: a7           [R] 1 tuple: <address>*1
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 2: aa ---- [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 3: a3       [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 4: a4 ------ [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 5: a5         [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 6: a6 -------- [-] 5 list: <address>*5, <address>*5, <address>*5...
 7: a7           [-] 1 list: <address>*5
 8: a8 ---------- [R] 1 tuple: <address>*1
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 2: aa ---- [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 3: a3       [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 4: a4 ------ [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 5: a5         [-] 5 list: <address>*5, <address>*5, <address>*5, <address>*5...
 6: a6 -------- [-] 5 list: <address>*5, <address>*5, <address>*5...
 7: a7           [-] 5 list: <address>*5, <address>*5, <address>*5...
 8: a8 ---------- [-] 5 list: <address>*5, <address>*5, <address>*5...
 9: a9             [-] 5 list: <address>*5, <address>*5, <address>*5...
<Type e.g. '_.more' for more.>
10: a10 ----------- [-] 5 list: <address>*5, <address>*5, <address>*5...
11: a11              [-] 5 list: <address>*5, <address>*5, <address>*5...
12: a12 ------------- [-] 5 list: <address>*5, <address>*5, <address>*5...
13: a13                [-] 5 list: <address>*5, <address>*5, <address>*5...
14: a14 --------------- [-] 5 list: <address>*5, <address>*5, <address>*5...
15: a15                  [-] 5 list: <address>*5, <address>*5, <address>*5...
16: a16 ----------------- [-] 5 list: <address>*5, <address>*5, <address>*5...
17: a17                    [+] 1 list: <address>*5
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [-] 1 list: <address>*1
 3: a3       [-] 1 list: <address>*1
 4: a4 ------ [-] 1 list: <address>*1
 5: a5         [-] 1 list: <address>*1
 6: a6 -------- [-] 1 list: <address>*1
 7: a7           [-] 1 list: <address>*1
 8: a8 ---------- [-] 1 list: <address>*1
 9: a9             [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
10: a10 ----------- [-] 1 list: <address>*1
11: a11              [-] 1 list: <address>*1
12: a12 ------------- [-] 1 list: <address>*1
13: a13                [-] 1 list: <address>*1
14: a14 --------------- [-] 1 list: <address>*1
15: a15                  [-] 1 list: <address>*1
16: a16 ----------------- [-] 1 list: <address>*1
17: a17                    [-] 1 list: <address>*1
18: a18 ------------------- [-] 1 list: <address>*1
19: a19                      [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
20: a20 --------------------- [-] 1 list: <address>*1
21: a21                        [-] 1 list: <address>*1
22: a22 ----------------------- [-] 1 list: <address>*1
23: a23                          [-] 1 list: <address>*1
24: a24 ------------------------- [-] 1 list: <address>*1
25: a25                            [-] 1 list: <address>*1
26: a26 --------------------------- [-] 1 list: <address>*1
27: a27                              [-] 1 list: <address>*1
28: a28 ----------------------------- [-] 1 list: <address>*1
29: a29                                [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
20: a20 --------------------- [-] 1 list: <address>*1
21: a21                        [-] 1 list: <address>*1
22: a22 ----------------------- [-] 1 list: <address>*1
23: a23                          [-] 1 list: <address>*1
24: a24 ------------------------- [-] 1 list: <address>*1
25: a25                            [-] 1 list: <address>*1
26: a26 --------------------------- [-] 1 list: <address>*1
27: a27                              [-] 1 list: <address>*1
28: a28 ----------------------------- [-] 1 list: <address>*1
29: a29                                [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
30: a30 ------------------------------- [-] 1 list: <address>*1
31: a31                                  [R] 1 tuple: <address>*1
30: a30 ------------------------------- [-] 1 list: <address>*1
31: a31                                  [R] 1 tuple: <address>*1
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [-] 1 list: <address>*1
 3: a3       [-] 1 list: <address>*1
 4: a4 ------ [-] 1 list: <address>*1
 5: a5         [-] 1 list: <address>*1
 6: a6 -------- [-] 1 list: <address>*1
 7: a7           [-] 1 list: <address>*1
 8: a8 ---------- [-] 1 list: <address>*1
 9: a9             [-] 1 list: <address>*1
<22 more lines. Type e.g. '_.more' for more.>
10: a10 ----------- [-] 1 list: <address>*1
11: a11              [-] 1 list: <address>*1
12: a12 ------------- [-] 1 list: <address>*1
13: a13                [-] 1 list: <address>*1
14: a14 --------------- [-] 1 list: <address>*1
15: a15                  [-] 1 list: <address>*1
16: a16 ----------------- [-] 1 list: <address>*1
17: a17                    [-] 1 list: <address>*1
18: a18 ------------------- [-] 1 list: <address>*1
19: a19                      [-] 1 list: <address>*1
<12 more lines. Type e.g. '_.more' for more.>
Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [-] 1 list: <address>*1
 3: a3       [-] 1 list: <address>*1
 4: a4 ------ [-] 1 list: <address>*1
 5: a5         [-] 1 list: <address>*1
 6: a6 -------- [-] 1 list: <address>*1
 7: a7           [-] 1 list: <address>*1
 8: a8 ---------- [-] 1 list: <address>*1
 9: a9             [-] 1 list: <address>*1
10: a10 ----------- [+] 1 list: <address>*1


Reference Pattern by <[dict of] class>.
 0: _ --- [-] 1 list: <address>*0
 1: a      [-] 1 list: <address>*1
 2: aa ---- [-] 1 list: <address>*1
 3: a3       [-] 1 list: <address>*1
 4: a4 ------ [-] 1 list: <address>*1
 5: a5         [-] 1 list: <address>*1
 6: a6 -------- [-] 1 list: <address>*1
 7: a7           [-] 1 list: <address>*1
 8: a8 ---------- [-] 1 list: <address>*1
 9: a9             [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
10: a10 ----------- [-] 1 list: <address>*1
11: a11              [-] 1 list: <address>*1
12: a12 ------------- [-] 1 list: <address>*1
13: a13                [-] 1 list: <address>*1
14: a14 --------------- [-] 1 list: <address>*1
15: a15                  [-] 1 list: <address>*1
16: a16 ----------------- [-] 1 list: <address>*1
17: a17                    [-] 1 list: <address>*1
18: a18 ------------------- [-] 1 list: <address>*1
19: a19                      [-] 1 list: <address>*1
<Type e.g. '_.more' for more.>
10: a10 ----------- [-] 1 list: <address>*1
11: a11              [-] 1 list: <address>*1
12: a12 ------------- [-] 1 list: <address>*1
13: a13                [-] 1 list: <address>*1
14: a14 --------------- [-] 1 list: <address>*1
15: a15                  [-] 1 list: <address>*1
16: a16 ----------------- [-] 1 list: <address>*1
17: a17                    [-] 1 list: <address>*1
18: a18 ------------------- [-] 1 list: <address>*1
19: a19                      [-] 1 list: <address>*1
20: a20 --------------------- [+] 1 list: <address>*1
""")

    def test_referrer_registration(self):
        import gc

        # The reference pattern should register itself as referrer target
        # so that after a gc, the rp target will still be included in the referrer target
        # Since the target is passed to referrers and update, it will still find the
        # referrers. It is an optimization issue: it should cover the referrers.
        # We test this by having two different-typed referrers
        # Accessing a referrer of the first one, then gc collecting, then checking that
        # the second one can be accessed without update: it was created automatically.
        # The test failed when not registering, but succeeded when registering was added.
        # It succeeds any case if no GC collection is made.

        dst = []
        a = [dst]
        aa = [a]
        b = (dst,)
        ba = [b]
        src = [aa, ba]
        rp = self.rp(dst, src)

        self.asis(rp._.theone, dst)
        gc.collect()
        self.asis(rp.aa.theone, aa)
        self.asis(rp.View.rg[b][0], ba)

    def test_some_more_advanced_usages(self):
        import gc

        # Test immediate dominators

        dst = []
        src = [dst]
        src.append([dst])
        rp = self.rp(dst, src, depth=10, imdom=1)
        self.asis(rp._.theone, dst)
        self.asis(rp.a.theone, src)

        # Test with mixed types

        # In particular, dict owned by an instance

        dst = []

        class A:
            pass
        a = A()
        a.dst = dst
        b = {'dst': dst}

        src = (a, b)

        gc.collect()
        rp = self.rp(dst, src, depth=10)
        rp.mod.View._is_clear_drg_enabled = 0  # Note Apr 19 2005
        self.asis(rp.a.theone, b)
        self.asis(rp.b.theone, a.__dict__)

        # Test that the dict is eventually automatically removed from dictowners -
        # First test that dictowners is nonzero

        ln = len(rp.mod.View.dict_ownership)
        self.assertTrue(ln > 0)

        del src
        del a
        mod = rp.mod
        rp.mod.View._is_clear_drg_enabled = 1
        del rp
        # It is cleared after GC

        gc.collect()
        lnnow = len(mod.View.dict_ownership)
        self.assertTrue(lnnow == 0)


class NewCase(TestCase):
    # Some new tests as they come up

    def test_reset(self):
        # Test the .reset() method

        dst = []
        a = [dst]
        b = [dst]
        src = [a, b]
        rp = self.rp(dst, src)
        self.aseq(rp.a, self.iso(a, b))

        b.pop()
        rp.reset()
        self.aseq(rp.a, self.iso(a))

    def test_paths(self):
        # Test the .paths() method

        dst = []
        a = [dst]+[None]*40     # Make order well-defined. Note May 2 2005.
        b = [dst]
        src = [a, b]
        rp = self.rp(dst, src)

        expected = """\
Paths from source 'a3' to target '_'.
 0: a3 [0] @ [0]
 1: aa [0]  @ [0]
 2: a  [0]   @ [0]
 3: _  [0]    = <1 list: <address>*0>
 4: aa [0]  @ [1]
 5: a  [1]   @ [0] -> #3"""

        self.aseq(str(rp.paths('a3')), expected)

        expected = expected[:expected.index('\n 4:')]

        # Test the andsets argument, given as a dict

        self.aseq(str(rp.paths('a3', andsets={'a': self.iso(a)})), expected)

        # Test the andsets argument, given as a list

        self.aseq(
            str(rp.paths('a3', andsets=[None, None, self.iso(a)])), expected)


def test_main(debug=0):
    support.run_unittest(RefPatCase, debug)
    support.run_unittest(NewCase, debug)


def test_leak():
    # Runs the tests in a loop and prints memory statistics,
    # to see if there are underlying low-level memory problems.
    # Requires Python to be compiled with debug support.
    from guppy.heapy.heapyc import xmemstats
    import gc
    import sys
    import time
    i = 0
    xmemstats()
    while 1:
        print('[%d]' % i, time.asctime())
        i += 1
        test_main()
        gc.collect()
        xmemstats()


if __name__ == "__main__":
    test_main()



from guppy.heapy.test import support
import sys
import unittest


class TestCase(support.TestCase):
    pass


class FirstCase(TestCase):
    def test_1(self):
        Spec = self.heapy.Spec
        TestEnv = Spec.mkTestEnv(Spec._Specification_)

        TestEnv.test_contains(Spec)


if __name__ == "__main__":
    support.run_unittest(FirstCase, 1)



class GSL_Error(Exception):
    pass


class TooManyErrors(GSL_Error):
    pass


class HadReportedError(GSL_Error):
    pass


class ReportedError(GSL_Error):
    pass


class UndefinedError(ReportedError):
    pass


class DuplicateError(ReportedError):
    pass


class CompositionError(GSL_Error):
    pass


class CoverageError(ReportedError):
    pass


class ConditionError(GSL_Error):
    pass



class _GLUECLAMP_:
    uniset_imports = ('UniSet', 'View', 'Path', 'RefPat')

    def _get_fa(self):
        us = self.UniSet
        us.out_reach_module_names = self.uniset_imports
        return us.fromargs



class Doc:  # base class
    def __mod__(self, other):
        other = self.mod.getdoc(other)
        return self.mapchildren(lambda x: x % other)

    def __rmod__(self, other):
        return self.mod.getdoc(other) % self

    def __str__(self):
        return self.getstr()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return 0

        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def shortest(self):
        return self.mapchildren(lambda x: x.shortest())


class Anon(Doc):
    def __init__(self, mod, obj):
        self.mod = mod
        self.obj = obj

    def getstr(self):
        return repr(self.obj)

    def mapchildren(self, f):
        return self


class Source(Doc):
    def __init__(self, mod, text):
        self.mod = mod
        self.text = text

    def getstr(self):
        return self.text

    def mapchildren(self, f):
        return self


class Attribute(Doc):
    def __init__(self, mod, obj, name):
        self.mod = mod
        self.obj = obj
        self.name = name

    def __mod__(self, other):
        if self.obj is other:
            return self.mod.rootattribute(other, self.name)
        return self.mapchildren(lambda x: x % other)

    def getstr(self):
        return '%s.%s' % (self.obj.getstr(), self.name)

    def mapchildren(self, f):
        return self.__class__(self.mod, f(self.obj), self.name)


class RootAttribute(Doc):
    def __init__(self, mod, obj, name):
        self.mod = mod
        self.obj = obj
        self.name = name

    def getstr(self):
        return '%s' % (self.name,)

    def mapchildren(self, f):
        return self


class BinaryOp(Doc):
    table = {
        'and': '&',
        'or': '|',
        'sub': '-',
        'mul': '*',
        'pow': '**',
        'lshift': '<<',
        'rshift': '>>',
    }

    def __init__(self, mod, op, a, b):
        self.mod = mod
        self.op = op
        self.a = a
        self.b = b

    def getstr(self):
        return '%s %s %s' % (self.a.getstr(),
                             self.table[self.op],
                             self.b.getstr())

    def mapchildren(self, f):
        return self.__class__(self.mod, self.op, f(self.a), f(self.b))


class UnaryOp(Doc):
    table = {
        'invert': '~',
        'neg': '-',
        'pos': '+',
    }

    def __init__(self, mod, op, a):
        self.mod = mod
        self.op = op
        self.a = a

    def getstr(self):
        return '%s %s' % (self.table[self.op], self.a.getstr())

    def mapchildren(self, f):
        return self.__class__(self.mod, self.op, f(self.a))


class CallFunc(Doc):
    def __init__(self, mod, obj, *args, **kwds):
        self.mod = mod
        self.obj = obj
        self.args = args
        self.kwds = kwds

    def getstr(self):
        return '%s(%s%s)' % (
            self.obj.getstr(),
            ', '.join([x.getstr() for x in self.args]),
            ', '.join(['%s=%s' % (k, v.getstr()) for k, v in list(self.kwds.items())]))

    def mapchildren(self, f):
        obj = f(self.obj)
        args = [f(a) for a in self.args]
        kwds = dict([(k, f(v)) for (k, v) in list(self.kwds.items())])
        return self.__class__(self.mod, obj, *args, **kwds)


class Multi(Doc):
    def __init__(self, mod, set):
        self.mod = mod
        self.str = '{%s}' % ', '.join([x.getstr() for x in set])
        self.set = set

    def getstr(self):
        return self.str

    def mapchildren(self, f):
        return self.__class__(self.mod, dict([(f(x), 1) for x in self.set]))

    def shortest(self):
        ls = None
        for a in self.set:
            a = a.shortest()
            l = len(a.getstr())
            if ls is None or l < ls:
                ls = l
                st = a
        return st


class Root(Doc):
    def __init__(self, mod, name='<root>'):
        self.mod = mod
        self.name = name

    def __call__(self, name):
        return self.__class__(self.mod, name)

    def mapchildren(self, f):
        return self

    def getstr(self):
        return self.name


class Tuple(Doc):
    def __init__(self, mod, *args):
        self.mod = mod
        self.args = args

    def mapchildren(self, f):
        return self.__class__(self.mod, *[f(x) for x in self.args])

    def getstr(self):
        x = '(%s)' % ', '.join([x.getstr() for x in self.args])
        if len(self.args) == 1:
            x = x[:-1]+',)'
        return x


class DocError(Exception):
    pass


class _GLUECLAMP_:

    def add_origin(self, obj, origin):
        o = getattr(obj, '_origin_', None)
        if o is None:
            obj._origin_ = origin
        else:
            obj._origin_ = self.multi(o, origin)
        return obj

    def add_wrapdict(self, obj, doc):
        wd = self.wrapdict
        o = wd.get(id(obj))
        if o is None:
            o = (obj, doc)
        else:
            o = (obj, self.multi(o[1], doc))
        wd[id(obj)] = o
        return obj

    def anon(self, obj):
        return Anon(self, obj)

    def attribute(self, obj, name):
        return Attribute(self, self.getdoc(obj), name)

    def binop(self, op, a, b):
        return BinaryOp(self, op, self.getdoc(a), self.getdoc(b))

    def callfunc(self, obj, *args, **kwds):
        getdoc = self.getdoc
        obj = getdoc(obj)
        args = [getdoc(a) for a in args]
        kwds = dict([(k, getdoc(v)) for (k, v) in list(kwds.items())])
        return CallFunc(self, obj, *args, **kwds)

    def getdoc(self, obj):
        if isinstance(obj, Doc):
            return obj
        w = getattr(obj, '_origin_', None)
        if isinstance(w, Doc):
            return w
        w = self.wrapdict.get(id(obj))
        if w is not None:
            return w[1]
        if isinstance(obj, tuple):
            return self.tuple(*obj)
        return self.anon(obj)

    def multi(self, a, b):
        a = self.getdoc(a)
        b = self.getdoc(b)
        if isinstance(a, Multi):
            set = a.set.copy()
            if len(set) > 4:
                return a
        else:
            set = {a: 1}
        if isinstance(b, Multi):
            set.update(b.set)
        else:
            set[b] = 1
        return Multi(self, set)

    def _get_root(self):
        return Root(self)

    def rootattribute(self, root, name):
        return RootAttribute(self, self.getdoc(root), name)

    def source(self, text):
        return Source(self, text)

    def tuple(self, *args):
        return Tuple(self, *[self.getdoc(x) for x in args])

    def unop(self, op, a):
        return UnaryOp(self, op, self.getdoc(a))

    def wrap(self, obj, doc):
        assert obj is not self._parent.UniSet.UniSet
        w = getattr(obj, '_derive_origin_', None)
        if w is not None:
            if getattr(w, 'im_self', None) is obj or isinstance(w, self._root.types.MethodType):
                obj = w(doc)
            elif w == 'ADD':
                obj = self.add_origin(obj, doc)
            else:
                raise DocError(
                    "Doc.wrap:  attribute '_derive_origin_' has invalid value")
        elif isinstance(obj, self._root.types.MethodType):
            obj = self.wrap_method(obj, doc)
        elif isinstance(obj, self._root.types.FunctionType):
            obj = self.wrap_function(obj, doc)
        else:
            obj = self.add_wrapdict(obj, doc)
        return obj

    def _get_wrapdict(self):
        return {}

    def wrap_function(mod, obj, doc):
        def f(*args, **kwds):
            r = obj(*args, **kwds)
            r = mod.wrap(r, mod.callfunc(doc, *args, **kwds))
            return r
        f._origin_ = doc
        return f

    def wrap_method(mod, obj, doc):
        im_func = obj.__func__

        def f(self, *args, **kwds):
            r = im_func(self, *args, **kwds)
            r = mod.wrap(r, mod.callfunc(doc, *args, **kwds))
            return r
        return mod._root.types.MethodType(f, obj.__self__)



import functools

from guppy.etc.RE_Rect import chooserects
from guppy.etc.IterPermute import iterpermute


class InfiniteError(Exception):
    pass


class WordsMemo:
    def __init__(self, re, ch):
        self.re = re
        self.ch = ch
        self.xs = {}
        self.N = 0

    def get_words_of_length(self, N):
        # Return a list of words of length up to N
        if N not in self.xs:
            self.xs[N] = self.re.get_words_of_length_memoized(N, self)
        return self.xs[N]

    def get_words_of_length_upto(self, N):
        # Return all words of length up to N, in the form
        # [(0, <list of words of length 0>),
        #  (1, <list of words of length 0>),
        #  ...]
        xsu = []
        for i in range(N+1):
            xs = self.get_words_of_length(i)
            if xs:
                xsu.append((i, xs))
        return xsu


REBASE = tuple


class RE(REBASE):
    # Regular expression nodes
    # The operators are choosen to be compatible with Pythonic standards:
    #   o sets               : using | for union
    #   o strings, sequences : using + for concatenation.
    #
    # This differs from mathematical presentations of regular
    # expressions where + is the union, but it seemed more important
    # to not confuse the Python usage.

    # There are also operators for closure x*, x+ that can not be
    # represented directly in Python expressions and these were choosen
    # to use a function call syntax.
    # The following table summarizes the operators.

    #   RE node expr    re lib          mathematical    name

    #   x + y           x y             x y             Concatenation
    #   x | y           x | y           x + y           Union
    #   x('*')          x*              x*              Kleene closure
    #   x('+')          x+              x+              Positive closure
    #   x('?')          x?

    _re_special = r'.^$*+?{}\[]|()'

    def __add__(a, b):
        if isinstance(b, RE):
            return concat(a, b)
        else:
            return Concatenation(a, Single(b))

    def __call__(a, *args, **kwds):
        if not kwds:
            if args == ('*',):
                return KleeneClosure(a)
            elif args == ('+',):
                return PositiveClosure(a)
            elif args == ('?',):
                return EpsilonOrOne(a)
        raise ValueError(
            "Argument to regular expression must be '*' or '+' or '?'")

    def __hash__(self):
        return hash((self._name, tuple(self)))

    def __eq__(a, b):
        return (a._name == b._name and
                tuple(a) == tuple(b))

    def __lt__(a, b):
        if a._name == b._name:
            return tuple(a) < tuple(b)
        else:
            return a._name < b._name

    def __or__(a, b):
        return Union(a, b)

    def get_num_closures(self):
        ns = 0
        for ch in self:
            ns += ch.get_num_closures()
        return ns

    def get_num_syms(self):
        ns = 0
        for ch in self:
            ns += ch.get_num_syms()
        return ns

    def get_sum_sym_lengths(self):
        ns = 0
        for ch in self:
            ns += ch.get_sum_sym_lengths()
        return ns

    def get_words_memo(self):
        ch = [x.get_words_memo() for x in self]
        return WordsMemo(self, ch)

    def get_words_of_length(self, N):
        xs = self.get_words_memo()
        return xs.get_words_of_length(N)

    def mapchildren(self, f):
        return self.__class__(*[f(x) for x in self])

    def regexpform(self):
        return self.mappedrepr(regexpname)

    def reversed(self):
        return self.mapchildren(lambda x: x.reversed())

    def rempretup(self):
        def f(x):
            if isinstance(x, Seq):
                if x is not Epsilon and isinstance(x[0], tuple):
                    ws = x[1:]
                    return Seq(*ws)
                else:
                    return x
            return x.mapchildren(f)

        return f(self)

    def seqatoms(self):
        sa = []
        self.apseqatoms(sa.append)
        return sa

    def sequni(self):
        d = {}
        us = []

        def ap(x):
            if x not in d:
                d[x] = 1
                us.append(x)
        self.apseq(ap)
        return Union(*us)

    def shform(self, conc=' '):
        r = self.mappedrepr(regexpname)
        if conc != ' ':
            r = conc.join(r.split(' '))
        return r

    def simplified(self, *a, **k):
        return self

    def simulform(self):
        def f(x):
            if x == '':
                return '()'
            return str(x)
        return self.mappedrepr(f)


def regexpname(s):
    if s == '':
        return '()'
    special = RE._re_special
    ren = []
    for c in str(s):
        if c in special+"', ":
            #c = r'\%s'%c
            c = ''
        ren.append(c)
    return ''.join(ren)


class Seq(RE):
    _priority = 0
    _name = 'Seq'

    def __new__(clas, *symbols):
        if not symbols:
            return Epsilon
        return REBASE.__new__(clas, symbols)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join(['%r' % (x,) for x in self]))

    def apseq(self, ap):
        ap(self)

    def apseqatoms(self, ap):
        for x in self:
            ap(Single(x))

    def get_num_closures(self):
        return 0

    def get_num_syms(self):
        return len(self)

    def get_sum_sym_lengths(self):
        s = 0
        for x in self:
            s += len(str(x))
        return s

    def get_words_memo(self):
        return WordsMemo(self, ())

    def get_words_of_length_memoized(self, N, memo):
        if N == len(self):
            return [self]
        else:
            return []

    def limited(self, N):
        return self

    def mappedrepr(self, f):
        if not self:
            return f('')
        return ' '.join(['%s' % (f(x),) for x in self])

    def reversed(self):
        r = list(self)
        r.reverse()
        return self.__class__(*r)

    def unionsplitted(self):
        return [self]


def Single(symbol):
    return REBASE.__new__(Seq, (symbol,))


Epsilon = REBASE.__new__(Seq, ())


def concat(*args):
    args = [x for x in args if x is not Epsilon]
    if len(args) < 2:
        if not args:
            return Epsilon
        return args[0]
    return REBASE.__new__(Concatenation, args)


class Concatenation(RE):
    _priority = 2
    _name = 'Concat'

    def __new__(clas, *args):
        if len(args) < 2:
            if not args:
                return Epsilon
            return args[0]
        return REBASE.__new__(clas, args)

    def __repr__(self):
        rs = []
        for ch in self:
            r = '%r' % (ch,)
            if ch._priority > self._priority:
                r = '(%s)' % (r,)
            rs.append(r)
        return ' + '.join(rs)

    def apseq(self, ap):
        uns = [x.sequni() for x in self]
        ixs = [0]*len(uns)
        while 1:
            xs = []
            for (i, us) in enumerate(uns):
                for x in us[ixs[i]]:
                    if x is not Epsilon:
                        xs.append(x)
            ap(Seq(*xs))
            j = 0
            for j, ix in enumerate(ixs):
                ix += 1
                if ix >= len(uns[j]):
                    ix = 0
                ixs[j] = ix
                if ix != 0:
                    break
            else:
                break

    def apseqatoms(self, ap):
        for x in self:
            x.apseqatoms(ap)

    def get_words_of_length_memoized(self, N, memo):
        chxs = []
        for ch in memo.ch:
            chxs.append(ch.get_words_of_length_upto(N))
        xs = []
        seen = {}

        def ads(xx, i, n):
            if i == len(chxs):
                if n == N:
                    for toconc in iterpermute(*xx):
                        conc = simple_Concatenation(toconc)
                        if conc not in seen:
                            xs.append(conc)
                            seen[conc] = 1
            else:
                for m, x in chxs[i]:
                    if n + m <= N:
                        ads(xx + [x], i + 1, n + m)

        ads([], 0, 0)
        return xs

    def limited(self, N):
        return Concatenation(*[x.limited(N) for x in self])

    def mappedrepr(self, f):
        rs = []
        for ch in self:
            r = ch.mappedrepr(f)
            if ch._priority > self._priority:
                r = '(%s)' % (r,)
            rs.append(r)
        return ' '.join(rs)

    def reversed(self):
        r = [x.reversed() for x in self]
        r.reverse()
        return self.__class__(*r)

    def simplified(self, *a, **k):
        conc = [x.simplified(*a, **k) for x in self]
        sa = []
        for c in conc:
            for a in c.seqatoms():
                sa.append(a)
        return simple_Concatenation(sa)

    def unionsplitted(self):
        runs = []
        uns = []
        for (i, x) in enumerate(self):
            us = x.unionsplitted()
            if len(us) > 1:
                uns.append((i, us))
        if not uns:
            return [self]
        ixs = [0]*len(uns)
        ch = list(self)
        while 1:
            xs = []
            i0 = 0
            for j, (i, us) in enumerate(uns):
                xs.extend(ch[i0:i])
                ix = ixs[j]
                xs.append(us[ix])
                i0 = i + 1
            xs.extend(ch[i0:])
            runs.append(concat(*xs))

            j = 0
            for j, ix in enumerate(ixs):
                ix += 1
                if ix >= len(uns[j][1]):
                    ix = 0
                ixs[j] = ix
                if ix != 0:
                    break
            else:
                return runs


class SimplifiedConcatenation(Concatenation):
    def simplified(self, *a, **k):
        return self


def conclosure(conc):
    # Simplification noted Mar 5 2005
    # Simplify ... b b* ... or ... b* b ... to ... b+ ...
    # conc is a sequence of regular expressions

    seen = {}
    nconc = []
    w0 = None
    for w in conc:
        if w0 is not None:
            if (w._name == '*' and      # Not isinstance(KleeneClosure), would catch PositiveClosure
                    w[0] == w0):
                w = PositiveClosure(w0)
            elif (w0._name == '*' and
                  w0[0] == w):
                w = PositiveClosure(w)
            else:
                if w0 is not None:
                    nconc.append(w0)
        w0 = w
    if w0 is not None:
        nconc.append(w0)
    return nconc


def simple_Concatenation(conc):
    if len(conc) > 1:
        conc0 = conc
        conc = conclosure(conc)
    nconc = []
    i = 0
    j = 0
    while i < len(conc):
        e = conc[i]
        if not isinstance(e, Seq):
            i += 1
            nconc.append(e)
            continue
        j = i
        while j < len(conc):
            if not isinstance(conc[j], Seq):
                break
            j += 1
        if j == i + 1:
            nconc.append(e)
        else:
            syms = []
            for k in range(i, j):
                e = conc[k]
                syms.extend(list(e))
            nconc.append(Seq(*syms))
        i = j
    if len(nconc) > 1:
        return Concatenation(*nconc)
    elif nconc:
        return nconc[0]
    else:
        return Epsilon


gauges = [
    lambda x:x.get_num_syms(),
    lambda x:x.get_num_closures(),
    lambda x:x.get_sum_sym_lengths()
]


def simpleunion(lines):
    choosen = chooserects(lines, gauges)
    have_epsilon = 0
    while 1:
        if len(choosen) == 1 and (choosen[0].width == 0 or len(choosen[0].lines) == 1):
            us = []
            for line in choosen[0].lines:
                if line:
                    us.append(line)
                else:
                    have_epsilon = 1
            break
        us = []
        for r in choosen:
            conc = r.get_common_part()
            olines = r.get_uncommons()
            u = simpleunion(olines)
            if u is not Epsilon:
                if r.dir == -1:
                    conc = [u]+conc
                else:
                    conc = conc + [u]
            if conc:
                us.append(conc)
            else:
                have_epsilon = 1
            assert not isinstance(us[-1], str)

        choosen = chooserects(us, gauges)

    if len(us) > 1:
        nus = [simple_Concatenation(line) for line in us]
        u = SimplifiedUnion(*nus)
    elif us:
        u = simple_Concatenation(us[0])
    else:
        u = None
    if have_epsilon:
        if u is not None:
            u = simple_EpsilonOrOne(u)
        else:
            u = Epsilon

    return u


class Union(RE):
    _priority = 3
    _name = 'Union'

    def __new__(clas, *args):
        return REBASE.__new__(clas, args)

    def __repr__(self):
        rs = []
        for ch in self:
            r = '%r' % (ch,)
            if ch._priority > self._priority:
                r = '(%s)' % r
            rs.append(r)
        return ' | '.join(rs)

    def apseq(self, ap):
        for c in self:
            c.apseq(ap)

    def apseqatoms(self, ap):
        for x in self:
            x.apseqatoms(ap)

    def get_words_of_length_memoized(self, N, memo):
        xs = []
        seen = {}
        for ch in memo.ch:
            for x in ch.get_words_of_length(N):
                if x not in seen:
                    seen[x] = 1
                    xs.append(x)
        return xs

    def limited(self, N):
        uni = [x.limited(N) for x in self]
        for i, x in enumerate(uni):
            if x is not self[i]:
                return self.__class__(*uni)
        return self

    def mappedrepr(self, f):
        rs = []
        for ch in self:
            r = '%s' % (ch.mappedrepr(f),)
            if ch._priority > self._priority:
                r = '(%s)' % r
            rs.append(r)
        return ' | '.join(rs)

    def simplified(self, args=None, *a, **k):
        if args is None:
            args = [x.simplified() for x in self.unionsplitted()]
            #args = [x for x in self.unionsplitted()]

        # Create a simplfied union
        # Assuming args are simplified, non-unions

        ch = [a.seqatoms() for a in args]
        return simpleunion(ch)

    def unionsplitted(self):
        us = []
        for x in self:
            us.extend(list(x.unionsplitted()))
        return us


class SimplifiedUnion(Union):
    def simplified(self, *a, **k):
        return self


class Called(RE):
    _priority = 1

    def __new__(clas, arg):
        return REBASE.__new__(clas, (arg,))

    def __repr__(self):
        ch = self[0]
        r = '%r' % (ch,)
        if ch._priority > self._priority:
            r = '(%s)' % r
        return "%s(%r)" % (r, self._name)

    def apseqatoms(self, ap):
        ap(self)

    def get_num_closures(self):
        return 1 + self[0].get_num_closures()

    def mappedrepr(self, f):
        ch = self[0]
        r = ch.mappedrepr(f)
        if (ch._priority > self._priority
                or isinstance(ch, Seq) and len(ch) > 1):
            r = '(%s)' % r
        return "%s%s" % (r, self._name)

    def simplified(self, *a, **k):
        return self.__class__(self[0].simplified(*a, **k))


class Closure(Called):
    def get_words_of_length_memoized(self, N, memo):
        if N == 0:
            return [Epsilon]
        if N == 1:
            return memo.ch[0].get_words_of_length(1)
        xs = []
        seen = {}
        for i in range(1, N):
            a = memo.get_words_of_length(i)
            b = memo.get_words_of_length(N-i)
            for ai in a:
                for bi in b:
                    aibi = simple_Concatenation((ai, bi))
                    if aibi not in seen:
                        xs.append(aibi)
                        seen[aibi] = 1
        for x in memo.ch[0].get_words_of_length(N):
            if x not in seen:
                xs.append(x)
                seen[x] = 1
        return xs

    def unionsplitted(self):
        return [self]


class KleeneClosure(Closure):
    _name = '*'

    def apseq(self, ap):
        raise InfiniteError(
            'apseq: Regular expression is infinite: contains a Kleene Closure')

    def limited(self, N):
        if N == 0:
            return Epsilon
        cl = self[0].limited(N)
        uni = []
        for i in range(N+1):
            toconc = [cl]*i
            uni.append(Concatenation(*toconc))
        return Union(*uni)

    def simplified(self, *a, **k):
        return simple_KleeneClosure(self[0].simplified(*a, **k))


def simple_KleeneClosure(x):
    # (b+)* -> b*
    if x._name == '+':
        return simple_KleeneClosure(x[0])
    return KleeneClosure(x)


class PositiveClosure(Closure):
    _name = '+'

    def apseq(self, ap):
        raise InfiniteError(
            'apseq: Regular expression is infinite: contains a Positive Closure')

    def apseqatoms(self, ap):
        self[0].apseqatoms(ap)
        simple_KleeneClosure(self[0]).apseqatoms(ap)

    def get_words_of_length_memoized(self, N, memo):
        if N <= 1:
            return memo.ch[0].get_words_of_length(N)
        return Closure.get_words_of_length_memoized(self, N, memo)

    def limited(self, N):
        a = self[0].limited(N)
        b = KleeneClosure(self[0]).limited(N)
        return Concatenation(a, b)


class EpsilonOrOne(Called):
    _name = '?'

    def apseq(self, ap):
        ap(Epsilon)
        self[0].apseq(ap)

    def get_words_of_length_memoized(self, N, memo):
        if N == 0:
            return [Epsilon]
        return memo.ch[0].get_words_of_length(N)

    def limited(self, N):
        x = self[0].limited(N)
        if x is not self[0]:
            self = self.__class__(x)
        return self

    def simplified(self, *a, **k):
        return simple_EpsilonOrOne(self[0].simplified(*a, **k))

    def unionsplitted(self):
        return [Epsilon] + list(self[0].unionsplitted())


def simple_EpsilonOrOne(x):
    # (a+)? -> a*

    if x._name == '+':
        return simple_KleeneClosure(x)

    # (a*)? -> a*
    if x._name == '*':
        return x

    return EpsilonOrOne(x)


class RegularSystem:

    def __init__(self, table, Start, final_states):
        self.table = table
        self.Start = Start
        self.Final = '358f0eca5c34bacdfbf6a8ac0ccf84bc'
        self.final_states = final_states

    def pp(self):

        def statename(state):
            try:
                name = self.names[state]
            except KeyError:
                name = str(state)
            return name

        def transname(trans):
            name = trans.simulform()
            if trans._priority > 1:
                name = '(%s)' % (name,)
            return name

        self.setup_names()

        X = self.X

        xs = [self.Start]+self.order
        xs.append(self.Final)
        for Xk in xs:
            if Xk not in X:
                continue
            print('%3s = ' % (statename(Xk),), end=' ')
            Tk = X[Xk]
            es = []
            for Xj in xs:
                if Xj in Tk:
                    es.append('%s %s' % (transname(Tk[Xj]), statename(Xj)))
            if es:
                print(' | '.join(es))
            else:
                print()

    def setup_equations(self):
        table = self.table
        final_states = self.final_states
        Final = self.Final
        self.X = X = {Final: {}}
        for Xi, transitions in list(table.items()):
            X[Xi] = Ti = {}
            for (symbol, Xj) in list(transitions.items()):
                Ti.setdefault(Xj, []).append(Single(symbol))
            for Xj, Aij in list(Ti.items()):
                if len(Aij) > 1:
                    Aij.sort()
                    Aij = Union(*Aij)
                else:
                    Aij = Aij[0]
                Ti[Xj] = Aij
            if Xi in final_states:
                Ti[Final] = Epsilon

    def setup_order(self):
        def dists(X, start):
            i = 0
            S = {start: i}
            news = [start]
            while news:
                oldnews = news
                news = []
                i += 1
                for s in oldnews:
                    if s not in X:
                        continue
                    for t in X[s]:
                        if t not in S:
                            news.append(t)
                            S[t] = i
            return S

        def start_distance(x):
            return start_dists[x]

        def sumt(f):
            memo = {}

            def g(x):
                if x in memo:
                    return memo[x]
                s = 0.0
                for y in X[x]:
                    s += f(y)
                memo[x] = s
                return s
            return g

        def cmp3(x, y):
            # Comparison for the sorting of equation solving order
            # First in list = solved last

            if x is y:
                return 0

            # Equations with more terms are resolved later
            c = len(X[y]) - len(X[x])
            if c:
                return c

            # The equations with terms more distant from start node will be resolved earlier
            i = 0
            while i < 10:  # 4 was enough with tests so far at Feb 24 2005
                try:
                    f = sumdists[i]
                except IndexError:
                    f = sumt(sumdists[i-1])
                    sumdists.append(f)
                c = f(x) - f(y)
                if c:
                    return c
                i += 1

            return (x > y) - (x < y)

        sumdists = [start_distance]
        X = self.X
        Start = self.Start
        Final = self.Final
        start_dists = dists(X, Start)
        order = [x for x in start_dists if x is not Start and x is not Final]
        order.sort(key=functools.cmp_to_key(cmp3))
        self.order = order

    def setup_names(self):
        try:
            self.order
        except AttributeError:
            self.setup_order()
        self.names = {}
        self.names[self.Start] = 'X0'

        for i, s in enumerate(self.order):
            self.names[s] = 'X%d' % (i+1)
        self.names[self.Final] = 'Final'

    def solve(self):
        # Set up equation system

        self.setup_equations()
        self.setup_order()

        X = self.X
        Start = self.Start
        Final = self.Final
        todo = list(self.order)

        # Solve equation system

        while todo:
            Xk = todo.pop()
            Tk = X[Xk]

            if Xk in Tk:
                # Recursive equation
                # Eliminate Akk Xk, using Adler's theorem
                # Given:
                # Xk = Ak0 X0 | ... Akk Xk |.. Akn Xkn
                # we get:
                # Xk = Akk* (Ak0 X0 | ... <no Xk> ... | Akn Xn)
                # which we evaluate to:
                # Xk = Bk0 X0 | ... Bkn Xn
                # where coefficients get the new values
                # Bki := Akk* Aki

                Akk = Tk[Xk]
                del Tk[Xk]

                AkkStar = Akk('*')
                for Xi, Aki in list(Tk.items()):
                    Bki = AkkStar + Aki
                    Tk[Xi] = Bki

            # Substitute Xk in each other equation in X
            # containing Xk, except eqv. Xk itself, which will not be used any more..

            del X[Xk]

            for Xj, Tj in list(X.items()):
                Bjk = Tj.get(Xk)
                if Bjk is None:
                    continue
                del Tj[Xk]
                for Xji, Tk_Xji in list(Tk.items()):
                    Cji = (Bjk + Tk_Xji)
                    Bji = Tj.get(Xji)
                    if Bji is not None:
                        Cji = Bji | Cji
                    Tj[Xji] = Cji

        # The equation system is now solved
        # The result is in Final term of Start equation

        return X[Start][Final]


Nothing = Union()


def SolveFSA(fsa):
    RS = RegularSystem(fsa.table, fsa.start_state, fsa.final_states)
    return RS.solve()



"""
Contains some experimental set constructions.
In the current state, not to be used by the faint-hearted.
It is likely due for some major refactoring and has probably
many bugs, it was a long time since I run the tests.
You don't get any other documentation than this right now.

"""


DEBUG = 1


def briefstr(x):
    try:
        return getattr(x, 'brief')
    except AttributeError:
        if isinstance(x, tuple):
            return '(%s)' % (','.join([briefstr(xi) for xi in x]))
        return str(x)


class SpecFamily:
    def __init__(self, unisetmod, specmod):
        self.specmod = specmod
        self.defrefining(unisetmod.Anything)
        self.disjoints -= [self]

    def __call__(self, arg):
        return self.Set(self, arg)

    def c_and(self, a, b):
        if isinstance(b, SpecFamily):
            return self.mod.fam_And._cons((a, b))
        else:
            return b.fam._and_ATOM(b, a)

    def _and_ID(self, a, b):
        env = self.specmod._static_test_env
        res = self.mod.mutnodeset()
        for bi in b.nodes:
            try:
                env.test_contains(a, bi, 'select contains')
            except Exception:  # TestError
                pass
            else:
                res.add(bi)
        return self.mod.fam_Identity._cons(self.mod.immnodeset(res))

    def c_contains(self, a, b):
        env = self.specmod._static_test_env
        try:
            return env.test_contains(a, b, 'Spec contains')
        except TestError:
            return False

    def c_getlimstr(self, a, max_len):
        x = a.brief
        if len(x) > max_len:
            x = x[:max_len-3]+'...'
        return x

    def c_get_brief(self, a):
        return '<%s(%s)>' % (self.__class__.__name__, briefstr(a.arg))


class ArgNamesFamily(SpecFamily):
    def __call__(self, *args):
        return self.Set(self, args)

    def getargnames(self, code):
        inspect = self.specmod._root.inspect
        (args, varargs, varkw) = inspect.getargs(code)
        if varargs is not None:
            args.append('*%s' % varargs)
        if varkw is not None:
            args.append('**%s' % varkw)
        return tuple(args)

    def func_argnames(self, f, args):
        try:
            code = f.__code__
            return self.getargnames(code) == args
        except AttributeError:
            return False

    def meth_argnames(self, m, args):
        try:
            f = m.__func__
            code = f.__code__
            return self.getargnames(code)[1:] == args
        except AttributeError:
            return False

    def c_contains(self, a, x):
        def func_argnames(f): return self.func_argnames(f, a.arg)
        def meth_argnames(m): return self.meth_argnames(m, a.arg)
        types = self.mod._root.types
        return (func_argnames(x) or
                meth_argnames(x) or
                (isinstance(x, type) and
                 (hasattr(x, '__init__') and
                  meth_argnames(x.__init__))) or
                (isinstance(x, type) and
                 (hasattr(x, '__init__') and
                  meth_argnames(x.__init__))) or
                (hasattr(x, '__call__') and
                 meth_argnames(x.__call__))
                )


class AttributeFamily(SpecFamily):
    def __call__(self, name, type=None):
        if type is None:
            type = self.specmod.any
        else:
            if not isinstance(type, self.mod.UniSet):
                type = self.c_from(type)
        return self.specotup((name, type))

    def c_test_contains(self, a, b, env):
        name, type = a.arg
        x = env.gengetattr(b, name)
        return type.test_contains(x, env)

    def c_get_brieflimstr(self, a):
        x = '<AttributeFamily(%r%%s>' % (a.arg[0],)
        if a.arg[1] is not None:
            x = x % (','+a.arg[1].brief)
        else:
            x = x % ''
        return x


class AdaptupleFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)
        self.add_export('new', lambda x: x.arg[0])

    def __call__(self, func, type):
        func, doc = self.specmod._exodoc(func)
        type = self.specmod.setcast(type)
        return self.specoargtup((func, type), (doc, type))

    def c_test_contains(self, a, b, env):
        func, type = a.arg
        if isinstance(b, tuple):
            try:
                x = func(*b)
            except Exception:
                return env.failed_exc_info('calling func failed')
        else:
            x = b
        return type.test_contains(x, env)


class CartesianProductFamily(SpecFamily):
    def __call__(self, *types):
        return self.Set(self, types)

    def c_domain(self, a):
        types = a.arg
        if len(types) == 2:
            return types[0]
        raise TypeError('Domain is defined on binary relations only')

    def c_get_examples(self, a, env):
        # We have to check length before calling iterpermute -
        # to give a reasonable error message examples are missing from some set.
        exs = []
        for i, ai in enumerate(a.arg):
            ex = list(env.get_examples(ai))
            if not ex:  # ???
                if not env.failed_coverage('cartesian product', ai, None, 'the argument #%d' % i):
                    return []
            exs.append(ex)

        return self.mod._root.guppy.etc.iterpermute(*exs)

    def c_len(self, a):
        l = 1
        for ai in a.arg:
            l = l * len(ai.arg)
        return l

    def c_test_contains(self, a, b, env):
        types = a.arg
        bs = tuple(b)
        if len(types) != len(bs):
            env.failed('cprod of length %d: argument has length %d' %
                       (len(types), len(bs)))
        else:
            for t, b in zip(types, bs):
                if not t.test_contains(b, env):
                    return False
            return True


class SequenceFamily(SpecFamily):
    def __call__(self, type):
        return self.specoarg(type)

    def c_test_contains(self, a, b, env):
        for x in b:
            if not a.arg.test_contains(x, env):
                return False
        return True

    def c_get_examples(self, a, env):
        for i in range(env.max_sequence_examples_length):
            for x in env.get_examples(self.specmod.cprod(*[a.arg]*i)):
                yield x


class MappingFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)
        self.add_export('arrow', lambda x: lambda f: self.c_arrow(x, f))

    def c_arrow(self, a, f):
        tf, ts, ret = a.arg
        return Arrow(ts, ret, f)

    def __call__(self, *args):
        ts = []
        as_ = []
        ret = None
        i = 0
        setcast = self.specmod.setcast
        while i < len(args):
            ai = args[i]
            if isinstance(ai, str):
                if ai == '->':
                    i += 1
                    if not i == len(args)-1:
                        raise SyntaxError(
                            "The '->' specifier must be next to last in the argument list")
                    ret = setcast(args[i])
                elif ai.endswith('='):
                    i += 1
                    t = setcast(args[i])
                    as_.append('%s=%args[%d]' % (ai[:-1], len(ts)))
                    ts.append(t)
                elif ai == '*':
                    i += 1
                    t = setcast(args[i])
                    as_.append('*args[%d]' % len(ts))
                    ts.append(self.specmod.sequence(t))
                else:
                    raise SyntaxError("Invalid argument specifier: %r" % ai)

            else:
                as_.append('args[%d]' % len(ts))
                ts.append(setcast(ai))
            i += 1

        fn = 'lambda f: lambda *args: f('+','.join(as_)+')'
        f = eval(fn)
        ts = self.specmod.cprod(*ts)
        if ret is None:
            ret = self.specmod.Anything
        return self.Set(self, (f, ts, ret))

    def c_test_contains(self, a, func, env):
        f, ts, ret = a.arg

        f = f(func)

        def g(env, args):
            try:
                r = f(*args)
            except Exception:  # TypeError, ValueError, AttributeError ... - what not
                raise
                #env.failed('Call exception')
            else:
                if ret is not None:
                    return env.test_contains(ret, r, "Mapping Return Type")
                else:
                    return True

        return env.forall(ts, g)

    def c_get_examples(self, a, env):
        f, ts, ret = a.arg

        return [lambda *args, **kwds: x for x in env.get_examples(ret)]


class PredicateFamily(SpecFamily):
    def __call__(self, arg, doc):
        return self.Set(self, (arg, doc))

    def c_test_contains(self, a, b, env):
        pred, doc = a.arg
        if not pred(env, b):
            return env.failed('pred: doc = %r; failed for element == %s' % (doc, env.name(b)))
        return True


class PowersetFamily(SpecFamily):
    def __init__(self, *args):
        SpecFamily.__init__(self, *args)
        self.add_export('union', lambda x: x.arg)

    def __call__(self, set):
        return self.specoarg(set)

    def c_get_examples(self, a, env):
        x = [self.specmod.Nothing, a.arg]
        try:
            y = env.get_examples(a.arg)
        except Exception:
            # xxx complain? We can always generate 2 examples..
            return x
        # xxx should we make all permutations?..
        return x

    def c_test_contains(self, a, b, env):
        if not b in self.specmod.set:
            env.failed('powerset.test_contains: not a set: %s' %
                       self.specmod.iso(b))
        set = a.arg
        return env.forall(b, lambda env, x: env.test_contains(set, x, 'powerset'), 'powerset')


class DocFamily(SpecFamily):
    def __call__(self, doc, type=None):
        if type is None:
            type = self.specmod.UniSet.NotNothing
        else:
            if not type in self.specmod.set:
                type = self.specmod.UniSet.convert(type)
        return self.Set(self, (doc, type))

    def c_test_contains(self, a, b, env):
        doc, type = a.arg
        return type.test_contains(b, env)


class RelOpFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)
        self.add_export('domain', lambda x: x.arg[0])
        self.add_export('range', lambda x: x.arg[2])
        # The memo dict keeps relops with the same domain, op, and range.
        # Primarily introduced to make equality work with default mechanism,
        # since different compiled func's compared differently even with
        # the same source.
        self.memo = {}

    def __call__(self, domain, op, range=None):
        domain = self.specmod.setcast(domain)
        if range is None:
            range = domain
        else:
            range = self.specmod.setcast(range)
        x = self.memo.get((domain, op, range))
        if x is None:
            if op in ('<', '<=', '==', '!=', '>', '>=', 'in', 'not in', 'is', 'is not'):
                func = eval('lambda x,y: x %s y' % op)
                func.name = op
            else:
                func = op
                func.name = func.__name__
            x = self.Set(self, (domain, func, range))
            self.memo[(domain, op, range)] = x
        return x

    def c_get_examples(self, a, env):
        # We have to check length before calling iterpermute -
        # to give a reasonable error message examples are missing from some set.
        dom = list(env.get_examples(a.domain))
        if not dom:
            failed_coverage('relation', a.domain, None, 'domain')
            return []

        ran = list(env.get_examples(a.range))
        if not ran:
            failed_coverage('relation', a.range, None, 'range')
            return []

        exs = []
        for ex in self.specmod._root.guppy.etc.iterpermute(dom, ran):
            if env.contains(a, ex):
                exs.append(ex)
        return exs

    def c_test_contains(self, a, b, env):
        d, op, r = a.arg
        if not op(*b):
            return env.failed()
        return True


class EqualsFamily(SpecFamily):
    def __call__(self, *args):
        if not args:
            return self.specmod.Nothing
        return self.specotup(args)

    def c_test_contains(self, a, b, env):
        if b in a.arg:
            return True
        return env.failed('equals')

    def c_get_examples(self, a, env):
        return a.arg


class ExampleFamily(SpecFamily):
    def __call__(self, set, *examples):
        return self.Set(self, (set, examples))

    def c_test_contains(self, a, b, env):
        set, examples = a.arg
        return set.test_contains(b, env)

    def c_get_examples(self, a, env):
        set, examples = a.arg
        # for e in examples:        env.test_contains(set, e, 'ExampleFamily.c_get_examples')
        return examples


class SynonymsFamily(SpecFamily):
    def __call__(self, *names):
        return self.Set(self, names)

    def c_test_contains(self, a, b, env):
        names = a.arg
        ms = self.specmod.UniSet.mutnodeset()
        for name in names:
            x = env.getattr(b, name)
            ms.add(x)
        if not len(ms):
            env.failed('Synonyms: no such names: %r' % names)
        if len(ms) > 1:
            env.failed('Synonyms: %d different nodes for names: %r' %
                       (len(ms), names))
        return True


class InstanceFamily(SpecFamily):
    def c_test_contains(self, a, b, env):
        if not isinstance(b, a.arg):
            env.failed('InstanceFamily: %s is not an instance of %s' % (
                self.specmod.iso(b), a.arg))
        return True

    def c_get_examples(self, a, env):
        return env.get_examples(self.specmod.setcast(a.arg))

    def _and_ID(self, a, b):
        return self.mod.fam_Identity(*[bi for bi in b.nodes if isinstance(bi, a.arg)])


class ExpressionPredicateFamily(SpecFamily):
    def __call__(self, names, expression):
        func = None
        return self.Set(self, (names, expression, func))

    def c_test_contains(self, a, b, env):
        names, expression, func = a.arg
        func = eval('lambda %s:%s' % (','.join(names), expression))
        d = {}
        for name in names:
            x = env.getattr(b, name)
            d[name] = x
        x = func(**d)
        if not x:
            env.failed('False expression: %s' % expression)
        return True


class ExpressionSetFamily(SpecFamily):
    def __call__(self, expression, *names):
        func = None
        return self.Set(self, (names, expression, func))

    def c_test_contains(self, a, b, env):
        names, expression, func = a.arg
        func = self.specmod.eval('lambda %s:(%s)' % (
            ','.join(('LE',)+tuple(names)), expression))
        d = {'LE': env.LE}
        for name in names:
            x = env.getattr(b, name)
            d[name] = x
        x = func(**d)
        return env.test_contains(x, b, 'expset(%s, %s)' % (expression, ','.join(names)))


class MatchesFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)
        self.sre = self.specmod._root.sre

    def __call__(self, regexp):
        return self.specoargtup(self.sre.compile(regexp), (regexp,))

    def c_test_contains(self, a, b, env):
        regexpobj = a.arg
        m = self.sre.match(regexpobj, b)
        if m is None:
            return env.failed('Did not match')
        return True


class RecurSelfFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)

        class RecurSpec(self.Set):
            __slots__ = 'func', 'recursion_level', 'recursion_limit'

        self.RecurSpec = RecurSpec

    def __call__(self, expr, recursion_limit=10):
        s = self.RecurSpec(self, expr)
        s.recursion_limit = recursion_limit
        s.recursion_level = 0
        return s

    def c_test_contains(self, a, b, env):
        try:
            func = a.func
        except AttributeError:
            expr = a.arg
            func = a.func = env.eval('lambda self:%s' % expr)
        s = func(self.specmod.Nothing)
        try:
            tf = env.test_contains(s, b, 'recur with Nothing, ok to fail')
            if not tf:
                raise TestError
        except Exception:  # TestError: eg for match, we got a TypeError..
            s = func(a)
            rl = a.recursion_level
            try:
                if rl >= a.recursion_limit:
                    return env.failed('recurself: recursion_level = %s' % a.recursion_limit)
                else:
                    a.recursion_level = rl + 1
                tf = env.test_contains(s, b, 'recur')
            finally:
                a.recursion_level = rl
        return tf


class Dummy:
    pass


class TupleformFamily(SpecFamily):
    def __call__(self, form, set):
        return self.specotup((form, set))

    def c_test_contains(self, a, b, env):
        form, set = a.arg

        bt = tuple(b)

        if len(bt) != len(form):
            return env.failed('tupleform: wrong length: got %d expected %d' % (len(bt), len(form)))

        ob = Dummy()
        for name, val in zip(form, bt):
            setattr(ob, name, val)

        return env.test_contains(set, ob, 'tupleform')


class Arrow:
    def __init__(self, category, source, target):
        self.category = category
        self.source = source
        self.target = target

    def of(self, other):
        # compose self <-- other
        # (Standard composition order)
        other = self.category.validate(other)
        return self.category.compose(self, other)

    def to(self, other):
        # compose self --> other
        # (Nonstandard composition order; sometimes seems more natural)
        other = self.category.validate(other)
        return self.category.compose(other, self)


class IdentityArrow(Arrow):
    def __init__(self, category, object):
        Arrow.__init__(self, category, object, object)

    def __call__(self, x):
        return x


class FunctionArrow(Arrow):
    def __init__(self, category, source, target, function):
        Arrow.__init__(self, category, source, target)
        self.__call__ = function


class RepresentationCategory:
    _derive_origin_ = None
    _origin_ = None

    def __init__(self, mod, spec):
        fam = mod.family(RepresentationObjectFamily)
        fam.init2(self, spec)
        self._fam = fam

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))

    def __getattr__(self, name):
        r = self._fam.getobject(name)
        self.__dict__[name] = r
        return r


class RepresentationObjectFamily(SpecFamily):
    def init2(self, cat, spec):
        self.cat = cat
        self.objects = {}
        self.specs = {}
        self.arrows = {}

        self.add_export('fromuniversal', self.fromuniversal)

        Set = self.Set

        class RepresentationObject(Set):
            __slots__ = 'to', 'from'

        self.Set = RepresentationObject

        class RepresentationCategorySpec(spec):
            def __init__(self, fam):
                self._fam = fam
                self._cat = fam.cat

            def __getattr__(self, name):
                if hasattr(self.__class__, '_get_%s' % name):
                    r = getattr(self, '_get_%s' % name)(self._fam.specmod)
                    self.__dict__[name] = r
                    return r
                raise AttributeError(name)

        self.spec = RepresentationCategorySpec(self)

    def getarrow(self, dom, cod):
        dc = (dom, cod)
        if dc in self.arrows:
            return self.arrows[dc]

        raise SpecError

    def getobject(self, name):
        if name in self.objects:
            return self.objects[name]
        normname = self.normoname(name)
        if normname in self.objects:
            self.objects[name] = self.objects[normname]
            return self.objects[normname]

        o = self(normname)
        self.objects[normname] = self.objects[name] = o
        return o

        raise SpecError('No such object: %r' % name)

    def getspec(self, obj):
        name = obj.arg
        if name in self.specs:
            return self.specs[name]

        gs = getattr(self.spec, '_get_spec_%s' % name, None)
        if gs is not None:
            sp = gs(self.specmod)
            self.specs[name] = sp
            return sp

        raise SpecError('getspec: No spec of %r' % name)

    def fromuniversal(self, target):
        # Find a most general arrow into 'target'
        # 1. Find all predefined arrows to target
        # 2. Make a category sum of the set of sources S of all such arrows
        # 3. Make a specification-set P for this sum, which is that:
        #    a. The element should be isomorphic to a tuple (O, A) where
        #    b. O, the tag, should be an object in S
        #    c. A, the value,  should be an element in O

        # Return an arrow taking this object to target.
        #
        # Arrows thus created are memoized.

        name = target.arg
        arrowname = '%s.fromuniversal' % name
        if arrowname in self.arrows:
            return self.arrows[arrowname]

        S = {}

        S[name] = IdentityArrow(self.cat, target)

        for an in dir(self.spec):
            if not (an.startswith('map_') and '_to_' in an):
                continue
            s, t = an.split('_to_')
            s = s[4:]
            if t == name:
                f = getattr(self.spec, an)
                S[s] = FunctionArrow(self.cat, getattr(self.cat, s), target,
                                     lambda x: f(self.specmod, x))

        def p(env, e):
            try:
                O, A = e
            except Exception:
                return env.failed('Not a sequence with length 2')
            fam = getattr(O, 'fam', None)
            if fam is not self:
                return env.failed('Tag is not an object of this family')
            name_ = getattr(O, 'arg', None)
            if name_ not in S:
                return env.failed('Object is not a source of this target')
            return env.test_contains(O, A, 'Value is not an element of this object')

        uniname = '%s.universal' % name
        P = self.specmod.predicate(p, 'Specification-set for %r' % uniname)
        self.specs[uniname] = P
        source = self(uniname)

        class AlphaSum(Arrow):
            def __init__(self, category, source, target, S):
                Arrow.__init__(self, category, source, target)

                self._S = S

            def __getitem__(self, x):   # Don't know what to call this unique arrow
                return self._S[self.source.fam.normoname(x)]

            def __call__(self, xxx_todo_changeme):
                (O, E) = xxx_todo_changeme
                return self[O.arg](E)

        arrow = AlphaSum(self.cat, source, target, S)
        self.arrows[arrowname] = arrow
        return arrow

    def normoname(self, on):
        isos = getattr(self.spec, '_isos_', None)
        if isos is not None:
            for l in self.spec._isos_:
                nn = None
                for n in l:
                    if n.isdigit():
                        nn = n
                    if n == on:
                        if nn is not None:
                            return nn
                        raise SpecError(
                            'No numeric object name corresponding to %r' % on)
        return on

    def __call__(self, name):
        r = self.specoarg(name)

        class To:
            __slots__ = 'source',

            def __init__(self, source):
                self.source = source

            def __call__(self, target):
                return self.source.fam.getarrow(self.source, target)

            def __getattr__(self, name):
                source = self.source
                fam = source.fam
                target = fam.getobject(name)
                return fam.getarrow(source, target)

        r.to = To(r)
        return r

    def c_test_contains(self, a, b, env):
        set = self.getspec(a)
        return env.test_contains(set, b, 'RepresentationObjectFamily: not in object specification')

    def c_get_examples(self, a, env):
        set = self.getspec(a)
        return env.get_examples(set)


class AbstractSetFamily(SpecFamily):
    def __init__(self, *a, **k):
        SpecFamily.__init__(self, *a, **k)

        class AbstractSet(self.Set):
            __slots__ = '_memo',

            def __init__(self, fam, arg):
                self.fam = fam
                self.arg = arg
                self._origin_ = None
                self._memo = {}
        self.Set = AbstractSet

    def __call__(self, arrow, set):
        r = self.specotup((arrow, set))
        r._memo = {}
        return r

    def c_and(self, a, b):
        if b.fam is self:
            aa, sa = a.arg
            ab, sb = b.arg
            if aa is ab:
                return self(aa, sa & sb)
        return b.fam._rand_ATOM(b, a)

    def c_getattr2(self, a, b):
        if b in a._memo:
            return a._memo[b]
        arrow, set = a.arg

        pararrow = arrow[b]
        srcset = pararrow.source

        def p(env, x):
            t = env.test_contains(
                srcset, x, 'AbstractSet: not in arrow.source')
            if t:
                e = pararrow(x)
                t = env.test_contains(
                    set, e, 'AbstractSet: not in argument set')
            return t

            return env.contains(set, y)

        s = self.specmod.predicate(p, 'Abstract set attribute: %r' % b)  # ...
        a._memo[b] = s
        return s

    def c_test_contains(self, a, b, env):
        arrow, set = a.arg
        t = env.test_contains(
            arrow.source, b, 'AbstractSet: not in arrow.source')
        if t:
            e = arrow(b)
            t = env.test_contains(set, e, 'AbstractSet: not in argument set')
        return t


class ElemFamily(SpecFamily):
    def __call__(self, func):
        return self.specoarg(func)

    def c_test_contains(self, a, b, env):
        func = a.arg
        s = func(b)
        return env.test_contains(s, b, 'ElemFamily')


class IntersectionFamily(SpecFamily):
    def __call__(self, domain, function):
        return self.specotup((domain, function))

    def c_test_contains(self, a, b, env):
        domain, function = a.arg
        return env.forall(domain, lambda env, x:
                          env.test_contains(function(x), b, 'Intersection'))


class SpecError(Exception):
    pass


class TestError(Exception):
    pass


class CoverageError(Exception):
    pass


class LocalEnv:
    _derive_origin_ = 'ADD'

    def __init__(self, mod, spec, nodoc=1):
        self._mod = mod
        self._spec = spec
        self._origin_ = None
        self.nodoc = nodoc

    def __getattr__(self, attribute_name):
        mod = self._mod
        f = getattr(self._spec, attribute_name)
        d = self._mod._load_names(
            mod._root.guppy.etc.Code.co_findloadednames(f.__code__))
        nf = mod._root.types.FunctionType(
            f.__code__,
            d,
            f.__name__,
            f.__defaults__,
            f.__closure__)
        r = nf(())
        self.__dict__[attribute_name] = r
        return r


class TestEnv:
    _derive_origin_ = 'ADD'
    iscomplete = False
    issilent = False
    # ie. (), cprod(x) cprod(x, y) are yielded by default
    max_sequence_examples_length = 2
    TestError = TestError

    def __init__(self, mod, Spec):
        self.mod = mod
        self.messages = []
        self.examples = {}
        if Spec is not None:
            self.spec = spec = Spec()
            try:
                lex = spec.LocalEnvExpr
            except AttributeError:
                lex = ''
            LE = LocalEnv(mod, lex)
            LE._OBJ_ = mod
            self.LE = LE

            self.topspec = self.eval(spec.GlueTypeExpr)

    def eval(self, expr):
        mod = self.mod
        types = mod._root.types
        if isinstance(expr, (str,)):
            func = self.mod.eval('lambda LE:(\n%s\n)' % expr)
            return func(self.LE)

        ls = []
        selfset = None

        names = list(expr.__dict__.keys())
        names.sort()

        for name in names:
            f = getattr(expr, name)
            try:
                co = f.__code__
            except AttributeError:
                continue
            if co.co_varnames[:co.co_argcount] == ('IN',):
                d = mod._load_names(
                    mod._root.guppy.etc.Code.co_findloadednames(co))
                #d = mod._load_names()

                nf = mod._root.types.FunctionType(
                    f.__code__,
                    d,
                    f.__name__,
                    f.__defaults__,
                    f.__closure__)
                s = nf(())
                if name == '_SELF_':
                    selfset = s
                else:
                    ls.append(mod.attr(name, s))
            else:
                raise SpecError('TestEnv.eval: invalid argument mode')
        # Constructing an AND in one sweep = faster
        # We assume they are not disjoint - which
        # would be determined by testing that we are going to do
        # (We know they are all attr's of different names here)

        # Except that selfset may perhaps be disjoint; but why care here
        #
        if selfset is not None:
            ls.append(selfset)

            # Alternatively: r = r & selfset afterwards,
            # but could be unnecessarily slow

        r = mod.UniSet.fam_And._cons(ls)
        return r

    def get_examples(self, collection):
        try:
            it = iter(collection)
        except TypeError:
            try:
                ex = self.examples[collection]
            except KeyError:
                if isinstance(collection, self.mod.UniSet.UniSet):
                    ex = collection.get_examples(self)
                else:
                    ex = list(collection)
            it = iter(ex)
        return it

    def getattr(self, obj, name):
        if '.' in name:
            x = obj
            for subname in name.split('.'):
                try:
                    x = getattr(x, subname)
                except Exception:
                    return self.failed('getattr: %s has no attribute %r' % (self.name(obj), name) +
                                       ',\nbecause %s has no attribute %r' % (self.name(x), subname))
        else:
            try:
                x = getattr(obj, name)
            except Exception:
                if DEBUG:
                    raise
                return self.failed('attr: %s has no attribute %r' % (self.name(obj), name))
        return x

    def gengetattr(self, obj, name_or_tuple):
        if isinstance(name_or_tuple, str):
            return self.getattr(obj, name_or_tuple)
        elif isinstance(name_or_tuple, tuple):
            if len(name_or_tuple) == 2 and name_or_tuple[0] is self.mod.quote:
                return name_or_tuple[1]
            else:
                return tuple([self.gengetattr(obj, nt) for nt in name_or_tuple])
        else:
            raise TypeError('gengetattr: I am picky, required string or tuple')

    def log(self, message):
        self.messages.append(message)

    def name(self, obj):
        if isinstance(obj, self.mod.UniSet.UniSet):
            return str(obj)
        else:
            return '%s' % self.mod.iso(obj)

    def name_coll(self, collection):
        return '%s' % collection

    def test(self, obj):
        self.get_obj_examples(obj)
        try:
            self.test_contains(self.topspec, obj, 'Top spec')
        except Exception:
            tvt = self.mod._root.sys.exc_info()
            self.dump_failure(tvt)

    def get_obj_examples(self, obj):
        ex = self.spec.GetExamples(self, obj)
        for e in ex:
            vs = e[1:]
            e = e[0]
            s = self.examples.setdefault(e, [])
            for v in vs:
                s.append(v)
                self.test_contains(e, v, 'testing example of set')

    def dump_failure(self, xxx_todo_changeme3, noraise=0):
        (type, value, traceback) = xxx_todo_changeme3
        list = []
        tb = traceback
        while tb is not None:
            f = tb.tb_frame
            if f.f_code is self.test_contains.__func__.__code__:
                list.append(f)
            tb = tb.tb_next
        for f in list:
            lo = f.f_locals
            print('a = %r' % (lo['a'],))
            print('b = %r' % (lo['b'],))
            print('message = ', lo['message'])
            print('-----')
        if noraise:
            self.mod._root.traceback.print_exception(type, value, traceback)
        else:
            raise

    def contains(self, a, b):
        try:
            x = self.test_contains(a, b, 'contains')
        except CoverageError:
            raise
        except Exception:
            return False
        return x

    def test_contains(self, a, b, message):
        return a.test_contains(b, self)

    def test_contains_not(self, a, b, message):
        try:
            a.test_contains(b, self)
        except CoverageError:
            raise
        except Exception:  # TestError: # well we axcept anything.. ok?
            return True
        else:
            return self.failed('test_contains_not, from: %s' % message)

    def failed(self, message=''):
        if not self.issilent:
            self.log('Failed:' + message)
            raise TestError(message)
        return False

    def failed_coverage(self, forwhat, collection, func, message):
        if collection is self.mod.Nothing:
            return True
        raise CoverageError('%s: no examples for collection = %s, message: %s' % (
            forwhat, collection, message))

    def failed_exc_info(self, message):
        exc_info = self.mod._root.sys.exc_info()
        type, value, traceback = exc_info
        if not self.issilent:
            self.log('Failed:' + message)
            raise type(value)
        return False

    def forall(self, collection, func, message=''):
        ex = self.get_examples(collection)
        n = 0
        for e in ex:
            if not func(self, e):
                return self.failed('forall: e = %s, from: %s' % (self.name(e), message))
            n += 1
        if not n:
            try:
                a = len(collection)
                if a > 0:
                    raise CoverageError
                # Otherwise ok, it was really an Nothing collection.
            except Exception:
                self.failed_coverage('forall', collection, func, message)

        return True

    def forall_pairs(self, collection, func, message=''):
        as_ = self.get_examples(collection)
        n = 0
        for a in as_:
            for b in self.get_examples(collection):
                if not func(self, a, b):
                    self.failed('forall_pairs: a = %s, b = %s, from: %s' % (
                        self.name(a), self.name(b), message))
                n += 1
        if not n:
            self.failed_coverage('forall_pairs', collection, func, message)
        return True

    def forall_triples(self, collection, func, message=''):
        as_ = self.get_examples(collection)
        n = 0
        for a in as_:
            for b in self.get_examples(collection):
                for c in self.get_examples(collection):
                    if not func(self, a, b, c):
                        self.failed('forall_triples: a = %s, b = %s, c=%s, from: %s' % (
                            self.name(a), self.name(b), self.name(c), message))
                    n += 1
        if not n:
            self.failed_coverage('forall_triples', collection, func, message)
        return True

    def forsome(self, collection, func, message=''):
        failures = []
        for i, x in enumerate(collection):
            try:
                b = func(x)
            except Exception:  # TestError:
                failures.append((i, self.mod._root.sys.exc_info()))
                continue
            if b:
                return True
        for (i, tvt) in failures:
            print('forsome: exception at collection[%d]:' % i)
            self.dump_failure(tvt, noraise=1)
        self.failed(message)


class _GLUECLAMP_:

    _chgable_ = '_loaded',
    _nowrap_ = ('_origin_', 'Doc', 'family', 'eval', 'setcast', 'compile', 'wrap_source_string',
                '_load_names', 'iso')

    _preload_ = '_hiding_tag_',

    # 'imports'

    _imports_ = (
        '_parent:Use',
        '_parent.Use:iso',
        '_parent.Use:Nothing',
    )

    def _get_UniSet(self): return self._parent.UniSet
    def _get_Doc(self): return self._parent.Doc
    def _get_View(self): return self._parent.View

    #

    def _get__origin_(self):
        return self.Doc.attribute(self.Doc.root, 'guppy.heapy.Spec')

    def _wrapattr_(self, obj, name):
        Doc = self.Doc
        if name == 'setof':
            pass
        try:
            obj = Doc.wrap(obj, Doc.attribute(self._origin_, name))
        except Doc.DocError:
            __import__('traceback').print_exc()
            print('no wrap:', name)
        return obj

    def _get_abstractset(self):
        return self.family(AbstractSetFamily)

    def _get_adaptuple(self):
        return self.family(AdaptupleFamily)

    def _get_any(self):
        return ~self.Nothing

    def _get_argnames(self):
        return self.family(ArgNamesFamily)

    def _get_attr(self):
        return self.family(AttributeFamily)

    def attrs(self, names, type=None):
        x = self.any
        for n in names:
            x &= self.attr(n, type)
        return x

    def _get_boolean(self):
        def p(env, x):
            try:
                if x:
                    pass
            except Exception:
                env.failed("boolean: 'if x' raised an exception")
            return True
        return self.predicate(p, 'boolean')

    def _get_callable(self):
        return self.predicate(lambda env, x: callable(x), 'callable')

    def _get_compile(self):
        token = self._root.token
        parser = self._root.parser
        symbol = self._root.symbol

        def mapchildren_id(node, f):
            return node

        def mapchildren_default(node, f):
            return [node[0]] + [f(n) for n in node[1:]]

        mapchildren_dispatch = dict(
            [(value, mapchildren_id) for value in range(token.N_TOKENS)])

        mapchildren_dispatch.update(dict([(value, mapchildren_default)
                                          for value, name in list(symbol.sym_name.items())]))

        def mapchildren(node, f):
            return mapchildren_dispatch[node[0]](node, f)

        def visitor(C):
            d = mapchildren_dispatch.copy()
            for value, name in list(symbol.sym_name.items()):
                if hasattr(C, name):
                    d[value] = getattr(C, name).__func__

            def _visit(node): return d[node[0]](node, _visit)
            return _visit

        def recover_source_node(node, l):
            for n in node[1:]:
                recover_source_dispatch[n[0]](n, l)

        def recover_source_token(node, l):
            if l and l[-1] != '(':
                l.append(' ')
            l.append(node[1])

        def recover_source_name(node, l):
            if l and l[-1] not in ('.', '('):
                l.append(' ')
            l.append(node[1])

        def recover_source_tight_left(node, l):
            l.append(node[1])

        def recover_source_lpar(node, l):
            if l and not (l[-1][-1:].isalnum() or l[-1] == '('):
                l.append(' ')
            l.append(node[1])

        recover_source_dispatch = dict([(value, recover_source_node)
                                        for value, name in list(symbol.sym_name.items())])
        recover_source_dispatch.update(
            dict([(value, recover_source_token) for value in range(token.N_TOKENS)]))

        recover_source_dispatch[token.NAME] = recover_source_name

        for tok in ('RPAR', 'LSQB', 'RSQB', 'COLON', 'COMMA', 'SEMI',
                    'DOT', 'LBRACE', 'RBRACE'):
            recover_source_dispatch[getattr(
                token, tok)] = recover_source_tight_left

        recover_source_dispatch[token.LPAR] = recover_source_lpar

        def recover_source(node):
            l = []
            recover_source_dispatch[node[0]](node, l)
            return ''.join(l)

        class wrap_lambdef:
            def test(node, f):
                # and_test ('or' and_test)* | lambdef
                if len(node) == 2 and node[1][0] == symbol.lambdef:
                    lsource = recover_source(node[1])
                    lnode = mapchildren(node[1], f)

                    return (
                        292, (293, (294, (295, (297, (298, (299, (300, (301,
                                                                        (302, (303, (304, (305, (1, 'wrap_source_string')),
                                                                                     (308, (7, '('), (317, (318, (292, lnode)), (12, ','),
                                                                                                      (318, (292, (293, (294, (295, (297, (298, (299, (300,
                                                                                                                                                       (301, (302, (303, (304, (305, (3, '%r' % lsource
                                                                                                                                                                                      )))))))))))))))), (8, ')'))))))))))))))

                return mapchildren(node, f)

        wrap_lambdef = visitor(wrap_lambdef)

        def compile_(source, filename, mode, *args):
            if mode != 'spec':
                return compile(source, filename, mode, *args)
            ast = parser.expr(source)
            node = parser.ast2tuple(ast, line_info=1)
            node = mapchildren(node, wrap_lambdef)

            ast = parser.tuple2ast(node)

            co = parser.compileast(ast)
            return co

        return compile_

    def _get_cprod(self):
        return self.family(CartesianProductFamily)

    def dictof(self, set=None):
        return self.UniSet.byclodo.dictof(set)

    def _get_doc(self):
        return self.family(DocFamily)

    def docof(self, set):
        doc = (set.doc % self._origin_).shortest()
        do = str(doc)
        print(do)
        return do

    def eval(self, expr, init=None, nodoc=0):
        if nodoc:
            mode = 'eval'
        else:
            mode = 'spec'
        co = self.compile(expr, '', mode)

        d = self._load_names(self._root.guppy.etc.Code.co_findloadednames(co))
        if init is not None:
            d = d.copy()
            exec(init, d)

        return eval(co, d)

    def _get_all_names(self):
        names = {'_root': 1}
        for x in _GLUECLAMP_.__dict__:
            if x.startswith('_get_'):
                x = x[5:]
            names[x] = 1
        return names

    def _load_names(self, names=None):
        if names is None:
            if '_loaded' not in self.__dict__:
                for x in self.all_names:
                    getattr(self, x)
                self._loaded = 1
        else:
            all = self.all_names
            for name in names:
                if name in all:
                    getattr(self, name)
        d = self.__dict__
        d['__builtins__'] = self._load_names.__globals__['__builtins__']
        return d

    def _get_eg(self):
        return self.family(ExampleFamily)

    def _get_elem(self):
        return self.family(ElemFamily)

    def _exodoc(self, expr):
        if expr in self.LE.string:
            r = self.eval(expr)
            return r, self.Doc.anon(expr)
        elif (expr in self.Type.Function and
              expr.__code__.co_name == '<lambda>' and
              expr.__code__.co_filename.startswith('<!SPECEVAL!>')):
            fn = expr.__code__.co_filename
            lines = fn.split('\n')
            lnum = expr.__code__.co_firstlineno
            inspect = self._root.inspect
            print(lines[lnum:])
            block = inspect.getblock(lines[lnum:])
            source = '\n'.join(block)
            return expr, self.Doc.getdoc(source)

        else:
            return expr, self.Doc.getdoc(expr)

    def _get_expred(self):
        return self.family(ExpressionPredicateFamily)

    def _get_expression(self):
        def p(env, x):
            try:
                eval('lambda : %s' % x)
            except SyntaxError:
                env.failed('Not a valid expression: %r' % x)
            return True
        return self.predicate(p, 'expression')

    def _get_expset(self):
        return self.family(ExpressionSetFamily)

    def _get_equals(self):
        return self.family(EqualsFamily)

    def family(self, fam, *args, **kwds):
        return self.UniSet.fam_mixin_argatom(fam, self, *args, **kwds)

    def _get__hiding_tag_(self):
        return self.UniSet._hiding_tag_

    def _get_instance(self):
        return self.family(InstanceFamily)

    def _get_intersection(self):
        return self.family(IntersectionFamily)

    def _get_LE(self):
        return LocalEnv(self, self._Specification_.LocalEnvExpr)

    def _get_SPLE(self):
        return self.LE

    def _get_mapping(self):
        return self.family(MappingFamily)

    def _get_matches(self):
        return self.family(MatchesFamily)

    def mkTestEnv(self, Spec):
        return TestEnv(self, Spec)

    def newtype(self, name):
        return self.predicate(lambda e, x: 1, name)  # xxx new family?

    def _get_nothing(self):
        return self.Use.Nothing

    def partition(self, set):
        return self.doc('partition',
                        self.union_equals(set) &
                        self.LE.nonempty &
                        self.set_of_disjoint_sets
                        )

    def _get_predicate(self):
        return self.family(PredicateFamily)

    def _get_powerset(self):
        return self.family(PowersetFamily)

    def _get_Anything(self):
        return self.UniSet.Anything

    def _get_quote(self):
        return []

    def _get_recurself(self):
        return self.family(RecurSelfFamily)

    def _get_relation(self):
        # relation is a representation category
        # We have the following representation objects:
        # 1. set of pairs with attribute dom, cod
        # 2. set of pairs with dom is cod
        # 3. fuop = func | op, string in ('==',..)
        # 4. (A, fuop), where A is a set
        # 5. (A, fuop, B), where A and B are sets
        #
        # A final object is 1.
        class RelationSpec:
            _isos_ = [
                ('1', 'paxb'),
                ('2', 'paxa'),
                ('3', 'defiop', 'fuop', ),
                ('4', 'defipair'),
                ('5', 'defitriple'),
            ]

            def _get_spec_1(self, e):
                return e.setof(e.cprod(e.Anything, e.Anything))

            def _get_spec_2(self, e):
                return (e.setof(e.cprod(e.Anything, e.Anything)))

            def _get_spec_3(self, e):
                return (e.boolean << (e.Anything, e.Anything) |
                        e.equals('<', '<=', '==', '!=', '>', '>=', 'in', 'not in', 'is', 'is not'))

            def _get_spec_4(self, e):
                return e.cprod(e.set, self._cat.fuop)

            def _get_spec_5(self, e):
                return e.cprod(e.set, self._cat.fuop, e.set)

            def _get__relop(self, e):
                return e.family(RelOpFamily)

            map_2_to_1 = id
            def map_3_to_4(self, e, fuop): return (e.Anything, fuop)
            def map_3_to_2(self, e, fuop): return self._relop(e.Anything, fuop)  # redundant
            def map_4_to_2(self, e, args): return self._relop(*args)
            def map_5_to_1(self, e, args): return self._relop(*args)

        return self.repcat(RelationSpec)

    def _get_relop(self):
        return self.relation.defiop.to.paxa

    def repcat(self, spec):
        return RepresentationCategory(self, spec)

    def _get_sequence(self):
        return self.family(SequenceFamily)

    def _get_set(self):
        return self.instance(self.UniSet.UniSet)

    def _get_set_of_disjoint_sets(self):
        def p(env, x):
            return env.forall_pairs(x, lambda env, a, b: a == b or a.disjoint(b),
                                    'a == b or a.disjoint(b)')
        return self.predicate(p, 'set of disjoint sets')

    def setcast(self, arg):
        if not isinstance(arg, self.UniSet.UniSet):
            arg = self.UniSet.convert(arg)
            arg = self.Doc.wrap(arg, self.Doc.callfunc(
                self.Doc.attribute(self._origin_, 'setcast'),
                arg))
        return arg

    def _get_setof(self):
        return self.powerset

    def _get__static_test_env(self):
        return self.mkTestEnv(None)

    def _get_synonyms(self):
        return self.family(SynonymsFamily)

    def _get_tupleform(self):
        return self.family(TupleformFamily)

    def _get_Type(self):
        return self.Use.Type

    def union_equals(self, set):
        return self.eg(self.attr('union', self.equals(set)),
                       self.powerset(set))

    def wrap_source_string(self, func, str):
        func._origin_ = self.Doc.source(str)
        return func



import importlib
import sys

autotests = (
    'dependencies',
    'gsl',
    'Classifiers',
    'heapyc',
    'ER',
    'OutputHandling',
    'Part',
    'Path',
    'RefPat',
    'UniSet',
    'View',
)

# These are not tested here

others = (
    'menuleak',
    'sf',
)


def test_main(debug=False):
    for name in autotests:
        testname = 'guppy.heapy.test.test_'+name
        try:
            del sys.modules[testname]
        except KeyError:
            pass
        x = importlib.import_module(testname)
        print('imported:', testname)
        f = x.test_main
        f(debug=debug)
        del sys.modules[testname]


if __name__ == '__main__':
    test_main()



import types
import weakref


# To restore the old-style class behavior that __getattr__ also affects special
# methods.
class _AttrProxy:
    _oh_proxied_classes = weakref.WeakSet({type, object})

    # Don'r rely on _oh_add_proxy_attr to prime us, some clients use
    # setup_printing to setup __repr__ function.
    def __repr__(self):
        return self.__getattr__('__repr__')()

    @classmethod
    def _oh_add_proxy_attr(cls, attr):
        if not attr.startswith('__') or not attr.endswith('__'):
            return
        for scls in cls.__mro__:
            if scls is object:
                break
            if attr in scls.__dict__:
                return
        if attr in ('__new__', '__init__', '__getattr__',
                    '__getattribute__', '__setattr__', '__delattr__',
                    '__str__',
                    ):
            return

        def closure(attr):
            def generated_function(self, *args, **kwds):
                func = self.__getattr__(attr)
                return func(*args, **kwds)

            return generated_function

        setattr(cls, attr, closure(attr))

    @classmethod
    def _oh_add_proxy_class(cls, base):
        if base in cls._oh_proxied_classes:
            return

        for scls in base.__mro__:
            cls._oh_proxied_classes.add(scls)
            for attr, val in scls.__dict__.items():
                if not isinstance(val, types.FunctionType):
                    continue

                cls._oh_add_proxy_attr(attr)


class OutputHandler:
    def __init__(self, mod, output_file):
        self.mod = mod
        self.output_file = output_file


class OutputBuffer:
    def __init__(self, mod, opts=None):
        self.mod = mod
        self.strio = mod._root.io.StringIO()

        if opts is None:
            opts = {}
        self.opts = opts

        self.lines = ['']
        self.line_no = 0

    def getopt(self, opt):
        return self.opts.get(opt)

    def getvalue(self):
        return '\n'.join(self.lines)

    def new_line(self):
        self.line_no += 1
        self.lines.append('')

    def write(self, s):
        lines = s.split('\n')
        for line in lines[:-1]:
            self.write_seg(line)
            self.new_line()
        self.write_seg(lines[-1])

    def write_seg(self, s):
        self.lines[self.line_no] += s


class AllPrinter(_AttrProxy):
    _oh_next_lineno = None

    def __init__(self, printer):
        self._oh_printer = printer
        self._hiding_tag_ = printer._hiding_tag_

    def __getattr__(self, attr):
        return self._oh_printer.getattr(self, attr)

    def __repr__(self):
        return self._oh_printer.get_str(self, True)

    def _oh_get_next_lineno(self):
        return 0

    def _oh_get_start_lineno(self):
        return 0

    def _oh_get_max_lines(self, max_lines):
        return None


class MorePrinter(_AttrProxy):
    _oh_next_lineno = None

    def __init__(self, printer, previous):
        self._oh_printer = printer
        self._oh_previous = previous
        self._hiding_tag_ = printer._hiding_tag_

    def __getattr__(self, attr):
        return self._oh_printer.getattr(self, attr)

    def _oh_get_next_lineno(self):
        next_lineno = self._oh_next_lineno
        if next_lineno is None:
            repr(self)
            next_lineno = self._oh_next_lineno
        return next_lineno

    def _oh_get_start_lineno(self):
        return self._oh_previous._oh_get_next_lineno()

    def _oh_get_max_lines(self, max_lines):
        return max_lines


class Printer:
    def __init__(self, mod, client, handler,
                 get_line_iter=None,
                 max_more_lines=None,
                 get_num_lines=None,
                 get_label=None,
                 get_row_header=None,
                 get_more_msg=None,
                 get_more_state_msg=None,
                 get_empty_msg=None,
                 stop_only_when_told=None
                 ):
        try:
            handler = handler.printer.handler
        except AttributeError:
            pass

        if get_line_iter is None:
            get_line_iter = handler._oh_get_line_iter
        if max_more_lines is None:
            max_more_lines = mod.max_more_lines

        self.mod = mod
        self._hiding_tag_ = mod._hiding_tag_
        self.client = client
        self.handler = handler
        self.get_line_iter = get_line_iter
        self.max_more_lines = max_more_lines
        if get_num_lines is None:
            get_num_lines = getattr(
                handler, '_oh_get_num_lines', None)
        if get_num_lines is not None:
            self.get_num_lines = get_num_lines
        if get_label is None:
            get_label = getattr(
                handler, '_oh_get_label', None)
        if get_label is not None:
            self.get_label = get_label
        if get_row_header is None:
            get_row_header = getattr(
                handler, '_oh_get_row_header', None)
        if get_row_header is not None:
            self.get_row_header = get_row_header
        if get_more_msg is None:
            get_more_msg = getattr(
                handler, '_oh_get_more_msg', None)
        if get_more_msg is not None:
            self.get_more_msg = get_more_msg
        if get_more_state_msg is None:
            get_more_state_msg = getattr(
                handler, '_oh_get_more_state_msg', None)
        if get_more_state_msg is not None:
            self.get_more_state_msg = get_more_state_msg
        if get_empty_msg is None:
            get_empty_msg = getattr(
                handler, '_oh_get_empty_msg', None)
        if get_empty_msg is not None:
            self.get_empty_msg = get_empty_msg
        self.stop_only_when_told = stop_only_when_told
        self.reset()

    def getattr(self, mp, attr):
        try:
            g = getattr(self, '_get_'+attr)
        except AttributeError:
            return getattr(self.client, attr)
        else:
            return g(mp)

    def line_at(self, idx):
        while idx >= len(self.lines_seen):
            try:
                li = next(self.line_iter)
            except StopIteration:
                raise IndexError
            else:
                if isinstance(li, tuple):
                    cmd, line = li
                    if cmd == 'STOP_AFTER':
                        self.stop_linenos[len(self.lines_seen)] = 1
                else:
                    line = li
                self.lines_seen.append(line)

        return self.lines_seen[idx]

    def lines_from(self, idx=0):
        line_iter = self.line_iter
        if line_iter is None:
            line_iter = self.line_iter = self.get_line_iter()
        while 1:
            try:
                yield self.line_at(idx)
            except IndexError:
                return
            idx += 1

    def _get_more(self, mp):
        return MorePrinter(self, mp)

    def _get_all(self, mp):
        return AllPrinter(self)

    def _oh_get_next_lineno(self):
        next_lineno = getattr(self, '_oh_next_lineno', None)
        if next_lineno is None:
            self.get_str_of_top()
            next_lineno = self._oh_next_lineno
        return next_lineno

    def _get_prev(self, mp):
        return mp._oh_previous

    def _oh_get_start_lineno(self):
        return 0

    def _oh_get_max_lines(self, max_lines):
        return max_lines

    def _get_top(self, mp):
        return self.client

    def _get___repr__(self, mp):
        return lambda: self.get_str(mp, False)

    def get_label(self):
        return None

    def get_row_header(self):
        return None

    def get_str_of_top(self):
        return self.get_str(self, True)

    def get_more_state_msg(self, start_lineno, end_lineno):
        num_lines = self.get_num_lines()
        if num_lines is None:
            of_num_lines = ''
        else:
            of_num_lines = ' of %d' % num_lines
        return "Lines %d..%d%s. " % (start_lineno, end_lineno, of_num_lines)

    def get_more_msg(self, start_lineno, end_lineno):
        state_msg = self.get_more_state_msg(start_lineno, end_lineno)
        return "<%sType e.g. '_.more' for more.>" % (state_msg)

    def get_empty_msg(self):
        return None

    def get_num_lines(self):
        return None

    def get_str(self, printer, is_top):
        def f():
            _hiding_tag_ = printer._hiding_tag_
            start_lineno = printer._oh_get_start_lineno()
            m_max_lines = printer._oh_get_max_lines(self.max_more_lines)
            ob = self.mod.output_buffer()
            it = self.lines_from(start_lineno)

            if is_top:
                msg = self.get_label()
                if msg is not None:
                    print(msg, file=ob)
                printer._oh_next_lineno = start_lineno

            try:
                nxt = next(it)
            except StopIteration:
                msg = self.get_empty_msg()
                if msg is not None:
                    print(msg, file=ob)
                printer._oh_next_lineno = start_lineno
            else:
                msg = self.get_row_header()
                if msg is not None:
                    print(msg, file=ob)

                it = self.mod.itertools.chain((nxt,), it)
                numlines = 0
                lineno = start_lineno
                for line in it:
                    if (m_max_lines is not None and numlines >= m_max_lines and
                            ((not self.stop_only_when_told) or self.stop_linenos.get(lineno-1))):
                        try:
                            self.line_at(lineno+1)
                        except IndexError:
                            print(line, file=ob)
                            lineno += 1
                            break
                        else:
                            print(self.get_more_msg(start_lineno, lineno-1), file=ob)
                            break
                    numlines += 1
                    print(line, file=ob)
                    lineno += 1
                printer._oh_next_lineno = lineno
            return ob.getvalue().rstrip()

        return self.mod.View.enter(lambda: f())

    def reset(self):
        self.lines_seen = []
        self.stop_linenos = {}
        self.line_iter = None


class _GLUECLAMP_:
    _chgable_ = 'output_file', 'max_more_lines',
    _preload_ = ('_hiding_tag_',)

    _imports_ = (
        '_parent:View',
        '_root:itertools',
    )

    max_more_lines = 10

    def _get__hiding_tag_(self): return self._parent.View._hiding_tag_
    def _get_output_file(self): return self._root.sys.stdout

    def more_printer(self, client, handler=None, **kwds):
        if handler is None:
            handler = client
        printer = Printer(self, client, handler, **kwds)
        _AttrProxy._oh_add_proxy_class(client.__class__)
        return MorePrinter(printer, printer)

    def output_buffer(self):
        return OutputBuffer(self)

    def output_handler(self, output_file=None):
        if output_file is None:
            output_file = self.output_file
        return OutputHandler(self, output_file)

    def setup_printing(self, client, handler=None, **kwds):
        if handler is None:
            handler = client
        more = self.more_printer(client, handler, **kwds)
        printer = more._oh_printer
        client.more = more
        client.all = more.all
        client.printer = printer

        def reprfunc(self):
            return self.printer.get_str_of_top()

        # Don't recreate, messes with relheap, and this is not
        # under _hiding_tag_
        cls = client.__class__
        if '__repr__' not in cls.__dict__:
            cls.__repr__ = reprfunc

    def _get_stdout(self): return self._root.sys.stdout



from guppy.heapy.test import support


class TestCase(support.TestCase):
    def setUp(self):
        support.TestCase.setUp(self)
        self.types = self.heapy.UniSet.types


class FirstCase(TestCase):
    def test_cal_hiding(self):
        V = self.View
        iso = self.iso

        # Tests that our workspace nodesets are hidden.

        a = []
        b = []
        as_ = iso(a)
        bs = iso(b)
        ab = as_ | bs
        # Make sure hiding is carried on with set operations
        self.assertTrue(ab.referrers.kind == self.types.FrameType)
        self.assertTrue((as_ | bs).referrers.kind == self.types.FrameType)
        self.assertTrue(V.referrers(iso(a)).kind == self.types.FrameType)
        self.assertTrue(V.referrers(iso(b)).kind == self.types.FrameType)

        # Test that we can still see nodesets created with other hiding_tag

        c = []
        chold = self.guppy.sets.immnodeset(
            [c],
            hiding_tag=[]     # Make sure we see it whether or not View uses default or not
        )
        cs = iso(c)
        del c

        self.assertTrue(cs.referrers.kind == type(
            self.guppy.sets.immnodeset()))

        h = self.View.heap()
        self.aseq(h & self.iso(self.Use), self.Use.Nothing)
        self.aseq(h & self.iso(self.Use.__dict__), self.Use.Nothing)

    def test_dominos(self):
        # Test dominos and domisize
        iso = self.iso
        x = []
        y = [x, []]
        z = [y]
        self.aseq(iso(y).dominos, iso(y, y[1]))
        self.aseq(iso(y).domisize, iso(y, y[1]).indisize)
        self.aseq(iso(z).dominos, iso(z))
        del y
        self.aseq(iso(z).dominos, iso(z, z[0], z[0][1]))

    def test_exports(self):
        # Test a few exports; the other defined in _unp_exports use the same mechanism
        iso = self.iso
        x = []
        y = [x, []]
        z = [y]
        p = iso(z).referents.indisize
        self.aseq(p, iso(y).indisize)

    def test_horizon(self):
        iso = self.iso
        h = self.View.horizon()
        x = []
        hn = h.news()
        self.aseq(hn, iso(x))
        del hn
        hn = h.news()
        self.aseq(hn, iso(x))
        del x, hn
        hn = h.news()
        self.aseq(hn, iso())

    def test_imdom(self):
        iso = self.iso
        x = []
        y = [x, []]
        z = [x, y]
        del x, y
        self.aseq(iso(z[0]).imdom, iso(z))

    def test_referents(self):
        iso = self.iso
        x = []
        y = [x, []]
        z = [y]
        self.aseq(iso(x).referents, iso())
        self.aseq(iso(y).referents, iso(x, y[1]))
        self.aseq(iso(z).referents, iso(y))
        self.aseq(iso(y, z).referents, iso(x, y, y[1]))

    def test_root(self):
        old_root = self.View.root
        self.assertIs(self.View.root, old_root)
        self.assertIs(self.Use.View.root, old_root)
        self.assertIs(self.View.hv.root, old_root)

        obj = object()
        self.View.root = obj
        self.assertIs(self.View.root, obj)
        self.assertIs(self.Use.View.root, obj)
        self.assertIs(self.View.hv.root, obj)

        self.View.root = old_root
        self.assertIs(self.View.root, old_root)
        self.assertIs(self.Use.View.root, old_root)
        self.assertIs(self.View.hv.root, old_root)


class GCCase(TestCase):
    def test_gc_mechanism(self):
        # Test the underlying GC system for having the properties that
        # support the method of automatic reclamation of nodegraphs
        # that is relied on in other tests. Failure here, dependent on
        # changed GC behaviour, would explain the other failures.

        from weakref import ref
        import gc

        class C:
            pass

        c = C()
        cbs = []

        def cb(wr):
            cbs.append(wr)

        wr = ref(c, cb)

        c.x = c
        gc.collect()
        strc = str(c)
        self.aseq(str(wr()), strc)
        self.asis(wr(), c)
        c = None
        self.aseq(str(wr()), strc)
        self.aseq(cbs, [])
        gc.collect()
        self.asis(wr(), None)
        self.aseq(cbs, [wr])

    def test_gc_hook(self):
        # Test the GC hook as implemented in View

        hos = []

        def ho():
            hos.append(1)

        import gc
        gc.collect()
        hook = self.heapy.View.gchook(ho)
        self.aseq(hos, [])
        gc.collect()
        self.aseq(hos, [1])
        hook = None
        gc.collect()
        self.aseq(hos, [1])

    def test_gc_drg(self):
        # Test automatic reclamation issues for dict owner nodegraph
        # This mechanism changed so the old test here is
        # not valid anymore
        # FIXME MAKE NEW TEST
        return

        import gc
        View = self.View
        hv = View.hv
        drg = View.nodegraph()

        def clear_drg():
            if drg.is_sorted:
                drg.clear()
            else:
                pass
        _clear_drg_hook = View.gchook(clear_drg)

        hv.update_dictowners(drg)
        gc.collect()
        lendrg = len(drg)
        self.assertTrue(lendrg > 0)    # Before any use, it will not be cleared
        # Now it is used by taking its length
        gc.collect()
        self.aseq(len(drg), 0)

        byclodo = hv.cli_clodo(drg, {})

        class C:
            pass
        c = C()

        byclodo.partition([c.__dict__])
        self.assertTrue(len(drg) > 0)
        gc.collect()
        self.assertTrue(len(drg) == 0)

    def test_gc_rg(self):
        # Test automatic reclamation issues for referrer nodegraph
        iso = self.iso
        immnodeset = self.guppy.sets.immnodeset
        self.View.is_rg_update_all = False
        import gc

        gc.collect()

        dst = []

        gc.collect()
        self.assertTrue(len(self.View.rg) == 0)

        # Test that rg is automatically updated with the set target(s)

        a = [dst]
        b = [dst]
        c = [a, b]

        class T:
            def __init__(self, set):
                self.set = set

        ta = T(iso(a))
        self.View.referrers_add_target(ta)
        tb = T(iso(b))
        self.View.referrers_add_target(tb)

        self.View.referrers(iso(c))
        self.assertTrue(c in immnodeset(self.View.rg[a]))
        self.assertTrue(c in immnodeset(self.View.rg[b]))

        # Test that a referrers target is automatically collected when not referenced

        tb = None
        gc.collect()
        self.View.referrers(iso(c))
        self.assertTrue(c in immnodeset(self.View.rg[a]))
        self.assertTrue(c not in immnodeset(self.View.rg[b]))

        # Test that adding a source automatically updates rg
        # even though domain of rg already includes its target
        # This requires gc collection

        d = [c]
        self.assertTrue(d not in self.View.referrers(iso(c)))
        gc.collect()
        self.assertTrue(d in self.View.referrers(iso(c)))


class AltHeapCase(TestCase):
    # Tests for support of alternative heap as implemented around 27 Oct 2005
    # and discussed in Notes at that time.

    def test_observation_containers(self):
        # Test the View.observation_containers method

        iso = self.iso
        idset = self.idset

        # These are to be included
        a = iso([], {})
        b = self.ImpSet.mutnodeset()
        c = self.View.observation_list()

        # These are not to be included - have another hiding tag

        excla = self.guppy.sets.mutnodeset()
        exclb = self.guppy.sets.immnodeset()

        # Get possible containers

        oc = idset(self.View.observation_containers())

        self.assertTrue(iso(a.nodes, b, c) <= oc)
        self.assertTrue(not (iso(excla, exclb) & oc))


class SpecialTypesCase(TestCase):
    # Some tests for special types

    def test_array(self):
        iso = self.iso
        import array
        a = array.array('b', b'asdf')
        iso(a).size


def test_main(debug=0):
    from guppy.heapy.Remote import off
    off()
    support.run_unittest(FirstCase, debug)
    support.run_unittest(GCCase, debug)
    support.run_unittest(AltHeapCase, debug)
    support.run_unittest(SpecialTypesCase, debug)


if __name__ == "__main__":
    test_main()



import os
import sys


class Target:
    def __init__(self, _hiding_tag_=None):
        self.wd = os.getcwd()
        self.pid = os.getpid()
        self.sys = sys
        self._hiding_tag_ = _hiding_tag_



class Doc2Latex:
    sizes = ('tiny', 'scriptsize', 'footnotesize', 'small',
             'normalsize', 'large', 'Large', 'LARGE', 'huge', 'Huge')

    def __init__(self, mod, doc, node):
        self.mod = mod
        self.doc = doc
        self.encoder = Encoder(mod)
        self.encode = self.encoder.encode
        self.node = node
        self.out = []
        self.ms = []
        self.mode = None
        self.cur_style = 'rm'
        self.cur_size = list(self.sizes).index('normalsize')
        self.document_lang = None
        self.document_title = None
        self.document_metas = []
        self.latex_list_nesting = 0
        self.latex_mode = 0
        self.noindent = 0
        self.authors = []

        node.accept(self)

    def _visit_children(self, node):
        E = self.mod.ReportedError
        for ch in node.children:
            try:
                ch.accept(self)
            except E:
                pass

    def abs_size(self, size, node):
        osize = self.cur_size
        si = size
        if si < 0:
            si = 0
        elif si >= len(self.sizes):
            si = len(self.sizes) - 1
        self.append('{\\%s ' % self.sizes[si])
        self.cur_size = si
        if self.cur_style != 'rm':
            self.style(self.cur_style, node)
        else:
            node.arg_accept(self)
        self.append('}')
        self.cur_size = osize

    def append(self, x):
        self.out.append(x)

    def changed_size(self, delta, node):
        self.abs_size(self.cur_size + delta, node)

    def error(self, msg, *args, **kwds):
        msg = 'Doc2Latex: ' + msg
        self.doc.env.error(msg, *args, **kwds)

    def get_latex(self):
        return ''.join(self.out)

    def no_children(self, node):
        if node.children:
            self.error('No children allowed for %r.' %
                       node.tag, node.children[0])

    def style(self, style, node):
        self.append('{\\%s ' % style)
        ostyle = self.cur_style
        self.cur_style = style
        node.arg_accept(self)
        self.cur_style = ostyle
        if style == 'em':
            self.append('\\/}')
        else:
            self.append('\\/}')

    def visit_a(self, node):
        pass

    def visit_author(self, node):
        self.authors.append(node.arg)
        self.no_children(node)

    def visit_big(self, node):
        self.changed_size(1, node)

    def visit_block(self, node):
        self._visit_children(node)

    def visit_blockquote(self, node):
        self.append('\\begin{quote}\n')
        self.latex_list_nesting += 1

        node.arg_accept(self)

        self.latex_list_nesting -= 1
        self.append('\\end{quote}\n')

    char_table = {
        'nbsp': '~',
    }

    def visit_char(self, node):
        char = node.arg.strip()
        c = self.char_table.get(char)
        if c is None:
            self.error('No such character: %r.' % char, node)
            c = char
        self.append(c)

    def visit_code(self, node):
        self.style('tt', node)

    def visit_comment(self, node):
        pass

    def visit_dd(self, node):
        self.ms.append('dd')
        step = 24
        ls = (self.ms.count('dd') + self.latex_list_nesting) * step
        self.append('{\\par \\noindent  \\leftskip = %d pt ' % ls)
        for i, v in enumerate(('i', 'ii', 'iii', 'iv', 'v', 'vi')[self.latex_list_nesting:]):
            self.append(' \\leftmargin%s = %d pt ' % (v, ls + (i + 1) * step))
        node.arg_accept(self)
        self.append('\\par}\n')
        self.ms.pop()

    def visit_default(self, node):
        self.error('I don\'t know what to generate for the tag %r.' %
                   node.tag, node)

    def visit_define(self, node):
        # xxx
        self._visit_children(node)

    def visit_dl(self, node):
        if self.ms and self.ms[-1] == 'dt':
            self.visit_dd(node)
        else:
            self.append('{\\par \\noindent\n')
            self._visit_children(node)
            self.append('\\par}\n')

    def visit_dt(self, node):
        self.ms.append('dt')
        self.append('{\\par \\pagebreak[%f] \\noindent \\hangindent = 12 pt \\hangafter = 1 \n' % (
                    3.4-0.1*len(self.ms),
                    ))
        node.arg_accept(self)
        self.append('\\par}\n')
        self.ms.pop()

    def visit_document(self, node):
        self._visit_children(node)

    def visit_document_lang(self, node):
        if self.document_lang is not None:
            self.error('Duplicate document lang directive.', node)
        self.document_lang = node

    def visit_document_title(self, node):
        if self.document_title is not None:
            self.error('Duplicate document title directive.', node)
        self.document_title = node

    def visit_exdefs(self, node):
        self.symplace = {}
        for ch in node.children:
            syms = [x.strip() for x in ch.arg.split(',')]
            for sym in syms:
                self.symplace[sym] = ch.tag

    def visit_em(self, node):
        self.style('em', node)

    def visit_enumerate(self, node):
        self.append('\\begin{enumerate}\n')
        for c in node.children:
            self.append('\\item ')
            c.accept(self)
        self.append('\\end{enumerate}\n')

    def visit_h0(self, node):
        # Not a html header,
        # we may treat this as 'new page' or chapter here
        # and some larger divisor in html.
        self.visit_hx(node)

    def visit_h1(self, node):
        self.visit_hx(node)

    def visit_h2(self, node):
        self.visit_hx(node)

    def visit_h3(self, node):
        self.visit_hx(node)

    def visit_h4(self, node):
        self.visit_hx(node)

    def visit_h5(self, node):
        self.visit_hx(node)

    def visit_h6(self, node):
        self.visit_hx(node)

    def visit_hx(self, node):
        n = int(node.tag[1:])
        if self.mode == 'man_page':
            self.append('{\\par \\pagebreak[%d] \\vskip %d pt \\noindent\n' % (
                [4, 3, 3, 2, 2, 1, 1][n],
                (12 - 2 * n)))
            self.abs_size(len(self.sizes) - n - 2, self.mod.node_of_taci(
                '', '',
                [self.mod.node_of_taci('strong', node.arg, node.children)]))
            self.append('\\par \\vskip %d pt\n} \\noindent\n' % (12 - 2 * n))
            self.noindent = 1
            # self.append('\\end{list}\n')
        else:
            self.append('\\%s{' % self.mod.section_table[n])
            node.arg_accept(self)
            self.append('}\n')

    def visit_itemize(self, node):
        self.append('\\begin{itemize}\n')
        self.latex_list_nesting += 1
        for c in node.children:
            self.append('\\item ')
            c.accept(self)

        self.latex_list_nesting -= 1
        self.append('\\end{itemize}\n')

    def visit_latex(self, node):
        self.latex_mode += 1
        node.arg_accept(self)
        self.latex_mode -= 1

    def visit_li(self, node):
        self.append('\\item ')
        node.arg_accept(self)

    def visit_link_to(self, node):
        # xxx
        name = node.arg
        self.append(' {\\em ')
        if not node.children:
            self.append(self.encode(name))
        else:
            self._visit_children(node)
        self.append('\\/}')

    def visit_link_to_extern(self, node):
        # xxx
        name = node.arg
        doc = node.children[0].arg
        children = node.children[1:]
        self.append(' {\\em ')
        if not children:
            self.append(self.encode(name))
        else:
            for ch in children:
                ch.accept(self)
        self.append('\\/}')

    def visit_link_to_local(self, node):
        # xxx
        name = node.arg
        self.append(' {\\em ')
        if not node.children:
            self.append(self.encode(name))
        else:
            self._visit_children(node)
        self.append('\\/}')

    def visit_link_to_unresolved(self, node):
        # xxx
        name = node.arg
        self.append(' {\\em ')
        if not node.children:
            self.append(self.encode(name))
        else:
            self._visit_children(node)
        self.append('\\/}')

    def visit_literal_block(self, node):
        self.append('{\\ttfamily \\raggedright \\noindent')
        self.encoder.literal_block = 1
        self.encoder.insert_none_breaking_blanks = 1

        node.arg_accept(self)
        self.encoder.literal_block = 0
        self.encoder.insert_none_breaking_blanks = 0
        self.append('}\n')

    def visit_lp(self, node):
        self.latex_mode += 1
        self.visit_paragraph(node)
        self.latex_mode -= 1

    def visit_man_page_mode(self, node):
        omode = self.mode
        self.mode = 'man_page'
        self._visit_children(node)
        self.mode = omode

    def visit_meta(self, node):
        self.document_metas.append(node)

    def visit_ol(self, node):
        self.append('\\begin{enumerate}\n')
        self._visit_children(node)
        self.append('\\end{enumerate}\n')

    def visit_p(self, node):
        self.visit_paragraph(node)

    def visit_paragraph(self, node):
        self.append('{\\par ')
        if self.noindent:
            self.append('\\parindent = 0 pt ')
            self.noindent = 0
        self.append('\n')
        node.arg_accept(self)
        self.append(' \\par}\n')

    def visit_pre(self, node):
        # I couldn't use Latex verbatim environment
        # since it didn't respected leftskip
        # so the environment became misplaced (within dd)

        text = node.arg.strip()
        if text:
            text += '\n'
        text = text + node.get_text()
        text = text.expandtabs()
        lines = text.split('\n')
        if lines and not lines[-1]:
            lines.pop()
        if not lines:
            return
        self.append('\\par\n')
        self.encoder.insert_none_breaking_blanks += 1
        self.encoder.literal += 1
        first = 1
        self.append('{\\tt{%s}}\n' % self.encode(lines[0]))
        for line in lines[1:]:
            self.append(
                '{ \\par \\parindent = 0 pt \\parskip = 0 pt \\tt{%s} }\n' %
                self.encode(line))
        self.encoder.insert_none_breaking_blanks -= 1
        self.encoder.literal -= 1
        self.append('\\par\n')

    def visit_small(self, node):
        self.changed_size(-1, node)

    def visit_spc_colonkind(self, node):
        self.append('~{\\bf :} ')

    def visit_spc_mapsto(self, node):
        self.append(' \\(\mapsto \\) ')

    def visit_string(self, node):
        self._visit_children(node)

    def visit_strong(self, node):
        self.style('bf', node)

    def visit_sub(self, node):
        self.append('\\raisebox{-.6ex}{')
        self.changed_size(-1, node)
        self.append('}')

    def visit_sup(self, node):
        self.append('\\raisebox{.6ex}{')
        self.changed_size(-1, node)
        self.append('}')

    def visit_symbol(self, node):
        self.visit_text(node)

    def visit_table(self, node):
        Table(self, node)

    def visit_text(self, node):
        if self.latex_mode:
            self.append(node.arg)
        else:
            text = node.arg
            text = self.encoder.encode(text)
            self.append(text)
        self._visit_children(node)

    def visit_to_document_only(self, node):
        self._visit_children(node)

    def visit_to_html_only(self, node):
        pass

    def visit_to_tester_only(self, node):
        pass

    def visit_tt(self, node):

        self.append('\\texttt{')
        self.encoder.literal = 1
        node.arg_accept(self)
        self.encoder.literal = 0
        self.append('}')

    def visit_ul(self, node):
        self.append('\\begin{itemize}\n')
        self._visit_children(node)
        self.append('\\end{itemize}\n')

    def visit_var(self, node):
        self.style('em', node)


class Table(Doc2Latex):
    many_hlines = 1  # Use extra many hlines.. looks good, a matter of taste.

    def __init__(self, d2l, node):
        self.d2l = d2l
        self.__dict__.update(d2l.__dict__)
        self.node = node
        self.out = []
        self.rows = []
        self.colwidth = None

        self._visit_children(node)

        maxcols = 0
        for row in self.rows:
            if len(row.columns) > maxcols:
                maxcols = len(row.columns)

        if not maxcols:
            return  # Empty table
        if self.colwidth is not None:
            if not len(self.colwidth) == maxcols:
                self.error("Wrong number of column width specifications (%d) vs\n"
                           "    max columns in table (%d)." % (
                               len(self.colwidth), maxcols),
                           node)
        else:
            self.colwidth = [1.0/maxcols]*maxcols
        ap = self.d2l.append
        ap('\n\\begin{longtable}[c]{|%s|}\n' % ('|'.join(['p{%.2g\\linewidth}' % cw
                                                          for cw in self.colwidth])))
        if self.many_hlines:
            ap('\\hline\n')
        for row in self.rows:
            for col in row.columns:
                ap(''.join(col.data))
                if col is row.columns[-1]:
                    if self.many_hlines:
                        ap('\\\\\n')
                        ap('\\hline\n')
                    else:
                        if row is not self.rows[-1]:
                            ap('\\\\\n')
                else:
                    ap('&\n')
            if row.is_head:
                ap('\\hline\n')
                ap('\\endhead\n')
        ap('\n\\end{longtable}\n')

    def visit_colgroup(self, node):
        colwidth = []

        for c in node.children:
            if c.tag != "col_width":
                self.error('Unrecognized colgroup option: %r' % c.tag, c)
            cg = c.arg
            if cg.endswith('%'):
                cg = cg[:-1]
                cg = float(cg)/100.0
            else:
                cg = float(cg)
            colwidth.append(cg)
        self.colwidth = colwidth

    def visit_options(self, node):
        pass

    def visit_thead(self, node):
        self._visit_children(node)
        self.rows[-1].is_head = 1

    def visit_tr(self, node):
        self.rows.append(Row(self, node))


class Row(Doc2Latex):
    is_head = 0

    def __init__(self, table, node):
        self.__dict__.update(table.__dict__)
        self.columns = []
        self._visit_children(node)

    def visit_td(self, node):
        self.columns.append(Column(self, node))

    def visit_th(self, node):
        self.columns.append(Column(self, node))


class Column(Doc2Latex):
    def __init__(self, row, node):
        self.__dict__.update(row.__dict__)
        self.data = []
        self.append = self.data.append
        node.arg_accept(self)


class Babel:
    """Language specifics for LaTeX."""
    # country code by a.schlock.
    # partly manually converted from iso and babel stuff, dialects and some
    _ISO639_TO_BABEL = {
        'no': 'norsk',  # XXX added by hand ( forget about nynorsk?)
        'gd': 'scottish',  # XXX added by hand
        'hu': 'magyar',  # XXX added by hand
        'pt': 'portuguese',  # XXX added by hand
        'sl': 'slovenian',
        'af': 'afrikaans',
        'bg': 'bulgarian',
        'br': 'breton',
        'ca': 'catalan',
        'cs': 'czech',
        'cy': 'welsh',
        'da': 'danish',
        'fr': 'french',
        # french, francais, canadien, acadian
        'de': 'ngerman',  # XXX rather than german
        # ngerman, naustrian, german, germanb, austrian
        'el': 'greek',
        'en': 'english',
        # english, USenglish, american, UKenglish, british, canadian
        'eo': 'esperanto',
        'es': 'spanish',
        'et': 'estonian',
        'eu': 'basque',
        'fi': 'finnish',
        'ga': 'irish',
        'gl': 'galician',
        'he': 'hebrew',
        'hr': 'croatian',
        'hu': 'hungarian',
        'is': 'icelandic',
        'it': 'italian',
        'la': 'latin',
        'nl': 'dutch',
        'pl': 'polish',
        'pt': 'portuguese',
        'ro': 'romanian',
        'ru': 'russian',
        'sk': 'slovak',
        'sr': 'serbian',
        'sv': 'swedish',
        'tr': 'turkish',
        'uk': 'ukrainian'
    }

    def __init__(self, mod):
        self.language = mod.language_code
        self.re = mod.re

        # pdflatex does not produce double quotes for ngerman in tt.
        self.double_quote_replacment = None
        if self.re.search('^de', self.language):
            #self.quotes = ("\"`", "\"'")
            self.quotes = ('{\\glqq}', '{\\grqq}')
            self.double_quote_replacment = "{\\dq}"
        else:
            self.quotes = ("``", "''")
        self.quote_index = 0

    def next_quote(self):
        q = self.quotes[self.quote_index]
        self.quote_index = (self.quote_index+1) % 2
        return q

    def quote_quotes(self, text):
        t = None
        for part in text.split('"'):
            if t == None:
                t = part
            else:
                t += self.next_quote() + part
        return t

    def double_quotes_in_tt(self, text):
        if not self.double_quote_replacment:
            return text
        return text.replace('"', self.double_quote_replacment)

    def get_language(self):
        if self.language in self._ISO639_TO_BABEL:
            return self._ISO639_TO_BABEL[self.language]
        else:
            # support dialects.
            l = self.language.split("_")[0]
            if l in self._ISO639_TO_BABEL:
                return self._ISO639_TO_BABEL[l]
        return None


class Encoder:
    literal_block = 0
    literal = 0
    mathmode = 0
    verbatim = 0
    insert_newline = 0
    mbox_newline = 0
    insert_none_breaking_blanks = 0

    latex_equivalents = {
        '\u00A0': '~',
        '\u2013': '{--}',
        '\u2014': '{---}',
        '\u2018': '`',
        '\u2019': '\'',
        '\u201A': ',',
        '\u201C': '``',
        '\u201D': '\'\'',
        '\u201E': ',,',
        '\u2020': '{\\dag}',
        '\u2021': '{\\ddag}',
        '\u2026': '{\\dots}',
        '\u2122': '{\\texttrademark}',
        '\u21d4': '{$\\Leftrightarrow$}',
    }

    def __init__(self, mod):
        self.mod = mod
        self.re = mod.re
        self.babel = Babel(mod)
        self.font_encoding = mod.font_encoding
        self.latex_encoding = self.to_latex_encoding(mod.output_encoding)

    def to_latex_encoding(self, docutils_encoding):
        """
        Translate docutils encoding name into latex's.

        Default fallback method is remove "-" and "_" chars from docutils_encoding.

        """
        tr = {"iso-8859-1": "latin1",     # west european
              "iso-8859-2": "latin2",     # east european
              "iso-8859-3": "latin3",     # esperanto, maltese
              "iso-8859-4": "latin4",     # north european,scandinavian, baltic
              "iso-8859-5": "iso88595",   # cyrillic (ISO)
              "iso-8859-9": "latin5",     # turkish
              "iso-8859-15": "latin9",    # latin9, update to latin1.
              "mac_cyrillic": "maccyr",   # cyrillic (on Mac)
              "windows-1251": "cp1251",   # cyrillic (on Windows)
              "koi8-r": "koi8-r",         # cyrillic (Russian)
              "koi8-u": "koi8-u",         # cyrillic (Ukrainian)
              "windows-1250": "cp1250",   #
              "windows-1252": "cp1252",   #
              "us-ascii": "ascii",        # ASCII (US)
              # unmatched encodings
              # "": "applemac",
              # "": "ansinew",  # windows 3.1 ansi
              # "": "ascii",    # ASCII encoding for the range 32--127.
              # "": "cp437",    # dos latine us
              # "": "cp850",    # dos latin 1
              # "": "cp852",    # dos latin 2
              # "": "decmulti",
              # "": "latin10",
              # "iso-8859-6": ""   # arabic
              # "iso-8859-7": ""   # greek
              # "iso-8859-8": ""   # hebrew
              # "iso-8859-10": ""   # latin6, more complete iso-8859-4
              }
        if docutils_encoding.lower() in tr:
            return tr[docutils_encoding.lower()]
        return docutils_encoding.translate(str.maketrans("", "", "_-")).lower()

    def unicode_to_latex(self, text):
        # see LaTeX codec
        # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/252124
        # Only some special chracters are translated, for documents with many
        # utf-8 chars one should use the LaTeX unicode package.
        for uchar in list(self.latex_equivalents.keys()):
            text = text.replace(uchar, self.latex_equivalents[uchar])
        return text

    def encode(self, text):
        """
        Encode special characters in `text` & return.
            # $ % & ~ _ ^ \ { }
        Escaping with a backslash does not help with backslashes, ~ and ^.

            < > are only available in math-mode or tt font. (really ?)
            $ starts math- mode.
        AND quotes:

        """
        if self.verbatim:
            return text
        # compile the regexps once. do it here so one can see them.
        #
        # first the braces.
        if 'encode_re_braces' not in self.__dict__:
            self.encode_re_braces = self.re.compile(r'([{}])')
        text = self.encode_re_braces.sub(r'{\\\1}', text)
        if 'encode_re_bslash' not in self.__dict__:
            # find backslash: except in the form '{\{}' or '{\}}'.
            self.encode_re_bslash = self.re.compile(r'(?<!{)(\\)(?![{}]})')
        # then the backslash: except in the form from line above:
        # either '{\{}' or '{\}}'.
        text = self.encode_re_bslash.sub(r'{\\textbackslash}', text)

        # then dollar
        text = text.replace("$", '{\\$}')
        if not (self.literal_block or self.literal or self.mathmode):
            # the vertical bar: in mathmode |,\vert or \mid
            #   in textmode \textbar
            text = text.replace("|", '{\\textbar}')
            text = text.replace("<", '{\\textless}')
            text = text.replace(">", '{\\textgreater}')
        # then
        text = text.replace("&", '{\\&}')
        # the ^:
        # * verb|^| does not work in mbox.
        # * mathmode has wedge. hat{~} would also work.
        # text = text.replace("^", '{\\ensuremath{^\\wedge}}')
        text = text.replace("^", '{\\textasciicircum}')
        text = text.replace("%", '{\\%}')
        text = text.replace("#", '{\\#}')
        text = text.replace("~", '{\\textasciitilde}')
        # Separate compound characters, e.g. "--" to "-{}-".  (The
        # actual separation is done later; see below.)
        separate_chars = '-'
        if self.literal_block or self.literal:
            # In monospace-font, we also separate ",,", "``" and "''"
            # and some other characters which can't occur in
            # non-literal text.
            separate_chars += ',`\'"<>'
            # pdflatex does not produce doublequotes for ngerman.
            text = self.babel.double_quotes_in_tt(text)
            if self.font_encoding == 'OT1':
                # We're using OT1 font-encoding and have to replace
                # underscore by underlined blank, because this has
                # correct width.
                text = text.replace('_', '{\\underline{ }}')
                # And the tt-backslash doesn't work in OT1, so we use
                # a mirrored slash.
                text = text.replace('\\textbackslash', '\\reflectbox{/}')
            else:
                text = text.replace('_', '{\\_}')
        else:
            text = self.babel.quote_quotes(text)
            text = text.replace("_", '{\\_}')
        for char in separate_chars * 2:
            # Do it twice ("* 2") becaues otherwise we would replace
            # "---" by "-{}--".
            text = text.replace(char + char, char + '{}' + char)
        if self.insert_newline or self.literal_block:
            # Insert a blank before the newline, to avoid
            # ! LaTeX Error: There's no line here to end.
            text = text.replace("\n", '~\\\\\n')
        elif self.mbox_newline:
            if self.literal_block:
                closings = "}" * len(self.literal_block_stack)
                openings = "".join(self.literal_block_stack)
            else:
                closings = ""
                openings = ""
            text = text.replace(
                "\n", "%s}\\\\\n\\mbox{%s" % (closings, openings))
        # lines starting with "[" give errors.
        text = text.replace('[', '{[}')
        if self.insert_none_breaking_blanks:
            text = text.replace(' ', '~')
        if self.latex_encoding != 'utf8':
            text = self.unicode_to_latex(text)
        return text


class _GLUECLAMP_:
    _imports_ = (
        '_parent:SpecNodes',
        '_parent.SpecNodes:node_of_taci',
        '_parent.Main:ReportedError',
        '_root:re',
    )

    font_encoding = ''
    double_quote_replacment = ''
    language_code = ''
    output_encoding = ''

    section_table = {
        0: 'part',
        1: 'chapter',
        2: 'section',
        3: 'subsection',
        4: 'subsubsection',
        5: 'paragraph',
        6: 'subparagraph'
    }

    def doc2text(self, doc, node):
        d2l = Doc2Latex(self, doc, node)
        return d2l.get_latex()

    def doc2filer(self, doc, node, name, dir, opts, IO):
        text = self.doc2text(doc, node)
        path = IO.path.join(dir, '%s.tex' % name)
        node = self.node_of_taci('write_file', path, [
                                 self.node_of_taci('text', text)])
        return node



import importlib

# This is to see that the total memory usage doesn't increase with time
# i.e. no leakage / link between consecutive usages of hsp.
# This will run for ever, to be monitored by the printout and some external monitor.


def t():
    from guppy import hsp
    while 1:
        import guppy.heapy.UniSet
        import gc
        importlib.reload(guppy.heapy.UniSet)
        hp = hsp()
        x = None
        x = hp.heap()
        print(x)
        gc.collect()
        print(x[0])
        print(x[1])
        print(x[2])
        gc.collect()
        print(x & dict)



from functools import reduce


def iterpermute(*args):
    args = [iter(a) for a in args]
    la = len(args)
    stopped = [0] * la
    lens = [0] * la
    bufs = [[] for i in range(la)]
    nexts = [None] * la
    n = 0
    while 1:
        anynew = 0
        for i in range(la):
            if stopped[i]:
                nextval = bufs[i][n % lens[i]]
            else:
                try:
                    nextval = next(args[i])
                except StopIteration:
                    if lens[i] == 0:
                        # raise ValueError, 'The iterator passed in arg %d did not return any item'%i
                        return
                    stopped[i] = 1
                    nextval = bufs[i][n % lens[i]]
                else:
                    anynew = 1
                    bufs[i].append(nextval)
                    lens[i] += 1
            nexts[i] = nextval
        if anynew:
            n += 1
            yield tuple(nexts)
        else:
            break

    wanted = reduce(lambda x, y: x*y, lens, 1)
    if n >= wanted:
        assert n == wanted
        return
    ixs = list(enumerate(lens))
    ixs.sort(key=lambda x: x[1])
    ixs = [ix for (ix, ln) in ixs]
    jxs = [0] * la
    seen = dict([(tuple([j % lens[i] for i in ixs]), 1)
                 for j in range(n)])

    while n < wanted:
        t = tuple([jxs[i] for i in ixs])
        if t not in seen:
            yield tuple([bufs[i][jxs[i]] for i in range(la)])
            n += 1

        for i in ixs:
            j = jxs[i]
            j = (j + 1) % lens[i]
            jxs[i] = j
            if j != 0:
                break


def test_iterpermute():
    import itertools
    repeat = itertools.repeat
    assert list(iterpermute()) == [()]
    assert list(iterpermute(repeat(1, 2))) == [(1,), (1,)]
    assert list(iterpermute(repeat(1, 1), repeat(2, 1))) == [(1, 2)]
    assert list(iterpermute(list(range(0, 2)), list(range(2, 3)))) == [
        (0, 2), (1, 2)]
    assert list(iterpermute(list(range(0, 2)), list(range(2, 4)))) == [
        (0, 2), (1, 3), (1, 2), (0, 3)]
    print(list(iterpermute(list(range(0, 2)), list(range(0, 3)))))
    print(list(iterpermute(list(range(0, 3)), list(range(0, 2)))))


if __name__ == '__main__':
    test_iterpermute()



""" Handles filing of data from low-level gsl filing and data records.
"""


class Filer:
    def __init__(self, mod, node):
        self.mod = mod
        self.writefile_envs = []
        self.writefile_names = {}

        node.accept(self)

    def visit_file(self, node):
        node.children_accept(self)

    visit_string = visit_file

    def visit_write_file(self, node):
        name = node.arg
        if name in self.writefile_names:
            raise SyntaxError('Duplicate file name: %r' % name)
        self.writefile_names[name] = node
        self.writefile_envs.append(WriteFile(self, node))

    def get_info(self):
        infos = []
        for e in self.writefile_envs:
            infos.append('write file: %s' % e.file_name)
        return '\n'.join(infos)

    def write(self):
        for e in self.writefile_envs:
            e.write()


class WriteFile:
    node_data = None
    node_mode = None

    def __init__(self, filer, node):
        self.filer = filer
        self.mod = mod = filer.mod
        self.node_file = node
        self.file_name = node.arg

        node.children_accept(self)
        if self.node_data is None:
            data = ''
        else:
            data = self.node_data.arg
        self.data = data
        if self.node_mode is None:
            mode = ''
        else:
            mode = self.node_mode.arg
        self.mode = mode

    def visit_text(self, node):
        self.set_single('node_data', node)

    def visit_end(self, node):
        self.set_single('node_end', node)

    def visit_mode(self, node):
        self.set_single('node_mode', node)

    def set_single(self, name, node):
        if getattr(self, name, None) is not None:
            raise SyntaxError('Duplicate %r at index %r' % (name, node.index))
        setattr(self, name, node)
        node.children_accept(self, 'no_node_expected')

    def write(self):
        IO = self.mod.IO
        if self.mod.backup_suffix:
            backup_name = self.file_name + self.mod.backup_suffix
            if IO.access(self.file_name, IO.R_OK | IO.W_OK):
                IO.rename(self.file_name, backup_name)

        IO.write_file(self.file_name, self.data)


class _GLUECLAMP_:
    _imports_ = (
        '_parent.FileIO:IO',
    )

    _setable_ = 'backup_suffix',

    # Files that are to be overwritten are renamed by
    # adding backup_suffix to the name. This is no substitute for a
    # versioning system but a last precaution, especially while I am
    # developing the system.
    # Set this to None to disable backuping.

    backup_suffix = ',gsl-backuped'

    syntax_gsl = '''
.tag writefile

'''

    def filer(self, node):
        return Filer(self, node)

    def _test_main_(self):
        IO = self.IO
        N = self._parent.SpecNodes
        tempdir = IO.mkdtemp()
        tempname = IO.path.join(tempdir, 'x')
        data = 'hello'
        try:
            X = '''
.write file: %s
..text
%s
..end
''' % (tempname, data)
            node = N.node_of_string(X)
            f = self.filer(node)
            assert f.get_info() == 'write file: %s' % tempname
            f.write()
            d = IO.read_file(tempname)
            assert d == data

            # Test multiple files and backup
            # And that we can do without ..data / ..end

            data2 = 'hello2\n'
            data3 = '\nhello3'
            X = '''
.write file: %s
..text
%s
.write file: %s
..text
%s
..end
''' % (tempname, data2, tempname+'.3', data3)

            node = N.node_of_string(X)
            f = self.filer(node)
            f.write()

            assert IO.read_file(tempname+self.backup_suffix) == data
            d = IO.read_file(tempname)
            assert d == data2
            assert IO.read_file(tempname+'.3') == data3

        finally:
            for name in IO.listdir(tempdir):
                IO.remove(IO.path.join(tempdir, name))
            IO.rmdir(tempdir)
